"""
Database proxy helper for X/84.
"""
import time


class DBProxy(object):
    """
    Provide dictionary-like object interface to a database. a database call,
    such as __len__() or keys() is issued as a command to the main engine,
    which spawns a thread to acquire a lock on the database and return the
    results via IPC pipe transfer.
    """

    def __init__(self, schema, table='unnamed'):
        """
        Arguments:
            schema: database key, to become basename of .sqlite3 files.
        """
        self.schema = schema
        self.table = table

    def proxy_iter(self, method, *args):
        """
        Iterable proxy for dictionary methods called over IPC pipe.
        """
        import x84.bbs.session
        event = 'db=%s' % (self.schema,)
        session = x84.bbs.session.getsession()
        session.flush_event(event)
        session.send_event(event, (self.table, method, args))
        data = session.read_event(event)
        assert data == (None, 'StartIteration'), (
            'iterable proxy used on non-iterable, %r' % (data,))
        data = session.read_event(event)
        while data != (None, StopIteration):
            yield data
            data = session.read_event(event)
        session.flush_event(event)

    def proxy_method(self, method, *args):
        """
        Proxy for dictionary methods called over IPC pipe.
        """
        import x84.bbs.session
        event = 'db-%s' % (self.schema,)
        session = x84.bbs.session.getsession()
        session.send_event(event, (self.table, method, args))
        return session.read_event(event)

    def acquire(self, blocking=True, stale=2.0):
        """
        Acquire a fine-grained BBS-global lock, blocking or non-blocking.

        When invoked with the blocking argument set to True (the default),
        block until the lock is acquired, and return True.

        When invoked with the blocking argument set to False, do not block.
        Returns False if lock is not acquired.

        If the engine has held the lock longer than ``stale`` seconds, the
        lock is granted anyway.
        """
        import x84.bbs.session
        event = 'lock-%s/%s' % (self.schema, self.table)
        session = x84.bbs.session.getsession()
        while True:
            session.send_event(event, ('acquire', stale))
            data = session.read_event(event)
            if data is True or not blocking:
                return data
            time.sleep(0.1)

    def release(self):
        """
        Release bbs-global lock on database.
        """
        import x84.bbs.session
        event = 'lock-%s/%s' % (self.schema, self.table)
        session = x84.bbs.session.getsession()
        return session.send_event(event, ('release', None))

    # pylint: disable=C0111
    #        Missing docstring
    def __contains__(self, key):
        return self.proxy_method('__contains__', key)
    __contains__.__doc__ = dict.__contains__.__doc__

    def __getitem__(self, key):
        return self.proxy_method('__getitem__', key)
    __getitem__.__doc__ = dict.__getitem__.__doc__

    def __setitem__(self, key, value):
        return self.proxy_method('__setitem__', key, value)
    __setitem__.__doc__ = dict.__setitem__.__doc__

    def __delitem__(self, key):
        return self.proxy_method('__delitem__', key)
    __delitem__.__doc__ = dict.__delitem__.__doc__

    def get(self, key, default=None):
        return self.proxy_method('get', key, default)
    get.__doc__ = dict.get.__doc__

    def has_key(self, key):
        return self.proxy_method('has_key', key)
    has_key.__doc__ = dict.has_key.__doc__

    def setdefault(self, key, value):
        return self.proxy_method('setdefault', key, value)
    setdefault.__doc__ = dict.setdefault.__doc__

    def update(self, *args):
        return self.proxy_method('update', *args)
    update.__doc__ = dict.update.__doc__

    def __len__(self):
        return self.proxy_method('__len__')
    __len__.__doc__ = dict.__len__.__doc__

    def values(self):
        return self.proxy_method('values')
    values.__doc__ = dict.values.__doc__

    def items(self):
        return self.proxy_method('items')
    items.__doc__ = dict.items.__doc__

    def iteritems(self):
        return self.proxy_iter('iteritems')
    iteritems.__doc__ = dict.iteritems.__doc__

    def iterkeys(self):
        return self.proxy_iter('iterkeys')
    iterkeys.__doc__ = dict.iterkeys.__doc__

    def itervalues(self):
        return self.proxy_iter('itervalues')
    itervalues.__doc__ = dict.itervalues.__doc__

    def keys(self):
        return self.proxy_method('keys')
    keys.__doc__ = dict.keys.__doc__

    def pop(self):
        return self.proxy_method('pop')
    pop.__doc__ = dict.pop.__doc__

    def popitem(self):
        return self.proxy_method('popitem')
    popitem.__doc__ = dict.popitem.__doc__



#!/usr/bin/env python
"""
Command-line launcher and main event loop for x/84
"""
# Please place _ALL_ Metadata in setup.py, except for a few bits
# which don't belong there right here. Don't include metadata in
# any other part of x/84, its a pita to maintain.
__author__ = "Johannes Lundberg, Jeffrey Quast"
__url__ = u'https://github.com/jquast/x84/'
__copyright__ = "Copyright 2012"
__credits__ = ["Wijnand Modderman-Lenstra", "zipe", "spidy", "Mercyful Fate"]
__license__ = 'ISC'


def main():
    """
    x84 main entry point. The system begins and ends here.

    Command line arguments to engine.py:
      --config= location of alternate configuration file
      --logger= location of alternate logging.ini file
    """
    # pylint: disable=R0914
    #        Too many local variables (19/15)
    import getopt
    import sys
    import os

    from x84.terminal import on_naws
    import x84.bbs.ini
    from x84.telnet import TelnetServer

    lookup_bbs = ('/etc/x84/default.ini',
                  os.path.expanduser('~/.x84/default.ini'))
    lookup_log = ('/etc/x84/logging.ini',
                  os.path.expanduser('~/.x84/logging.ini'))
    try:
        opts, tail = getopt.getopt(sys.argv[1:], u'', (
            'config=', 'logger=', 'help'))
    except getopt.GetoptError as err:
        sys.stderr.write('%s\n' % (err,))
        return 1
    for opt, arg in opts:
        if opt in ('--config',):
            lookup_bbs = (arg,)
        elif opt in ('--logger',):
            lookup_log = (arg,)
        elif opt in ('--help',):
            sys.stderr.write('Usage: \n%s [--config <filepath>] '
                             '[--logger <filepath>]\n' % (
                                 os.path.basename(sys.argv[0],)))
            sys.exit(1)
    if len(tail):
        sys.stderr.write('Unrecognized program arguments: %s\n' % (tail,))

    # load/create .ini files
    x84.bbs.ini.init(lookup_bbs, lookup_log)

    # init userbase pw encryption
    import x84.bbs.userbase
    x84.bbs.userbase.digestpw_init(
        x84.bbs.ini.CFG.get('system', 'password_digest'))

    # start telnet server
    telnetd = TelnetServer((
        x84.bbs.ini.CFG.get('telnet', 'addr'),
        x84.bbs.ini.CFG.getint('telnet', 'port'),),
        on_naws)

    # begin main event loop
    _loop(telnetd)


def _loop(telnetd):
    """
    Main event loop. Never returns.
    """
    # pylint: disable=R0912,R0914,R0915
    #         Too many local variables (24/15)
    import logging
    import select
    import socket
    import time
    import sys

    from x84.bbs.exception import Disconnected
    from x84.terminal import terminals, ConnectTelnet, unregister
    from x84.bbs.ini import CFG
    from x84.telnet import TelnetClient
    from x84.db import DBHandler

    logger = logging.getLogger()

    if sys.maxunicode == 65535:
        logger.warn('Python not built with wide unicode support!')
    logger.info('listening %s/tcp', telnetd.port)
    timeout_ipc = CFG.getint('system', 'timeout_ipc')
    fileno_telnetd = telnetd.server_socket.fileno()
    locks = dict()

    def lookup(client):
        """
        Given a telnet client, return a matching session
        of tuple (client, inp_queue, out_queue, lock).
        If no session matches a telnet client,
            (None, None, None, None) is returned.
        """
        for _sid, tty in terminals():
            if client == tty.client:
                return tty
        return None

    def telnet_recv(fds):
        """
        test all telnet clients with file descriptors in list fds for
        recv(). If any are disconnected, signal exit to subprocess,
        unregister the session (if any).
        """
        # read in any telnet input
        for client in (clt for (fno, clt) in telnetd.clients.iteritems()
                       if fno in fds):
            try:
                client.socket_recv()
            except Disconnected as err:
                logger.info('%s Connection Closed: %s.',
                            client.addrport(), err)
                tty = lookup(client)
                if tty is None:
                    # no session found, just de-activate this client
                    client.deactivate()
                else:
                    # signal exit to sub-process and shutdown
                    send_event, send_data = 'exception', Disconnected(err)
                    tty.iqueue.send((send_event, send_data))
                    unregister(tty)

    def telnet_send(recv_list):
        """
        test all telnet clients for send(). If any are disconnected,
        signal exit to subprocess, unregister the session, and return
        recv_list pruned of their file descriptors.
        """
        for sid, tty in terminals():
            if tty.client.send_ready():
                try:
                    tty.client.socket_send()
                except Disconnected as err:
                    logger.debug('%s Disconnected: %s.', sid, err)
                    # client.sock.fileno() can raised 'bad file descriptor',
                    # so, to remove it from the recv_list, reverse match by
                    # instance saved with its FD as a key for telnetd.clients!
                    for _fd, _client in telnetd.clients.items():
                        if tty.client == _client:
                            if _fd in recv_list:
                                recv_list.remove(_fd)
                            break
                    send_event, send_data = 'exception', Disconnected(err)
                    tty.iqueue.send((send_event, send_data,))
                    unregister(tty)
        return recv_list

    def accept():
        """
        accept new connection from telnetd.server_socket, and
        instantiate a new TelnetClient, registering it with
        dictionary telnetd.clients, and spawning an unmanaged
        thread for negotiating TERM.
        """
        try:
            sock, address_pair = telnetd.server_socket.accept()
            # busy signal
            if telnetd.client_count() > telnetd.MAX_CONNECTIONS:
                try:
                    sock.shutdown(socket.SHUT_RDWR)
                except socket.error:
                    pass
                sock.close()
                logger.error('refused new connect; maximum reached.')
                return
            client = TelnetClient(sock, address_pair, telnetd.on_naws)
            telnetd.clients[client.sock.fileno()] = client
            # spawn negotiation and process registration thread
            ConnectTelnet(client).start()
            logger.info('%s Connected.', client.addrport())
        except socket.error as err:
            logger.error('accept error %d:%s', err[0], err[1],)

    @timeout_alarm(timeout_ipc, False)   # REMOVEME
    def f_send_event(iqueue, event, data):
        """
        Send event to subprocess, signaling an alarm timeout if blocked.
        """
        iqueue.send((event, data))
        return True

    def send_input(client, iqueue):
        """
        Send tcp input to subprocess as 'input' event,
        signaling an alarm timeout and re-buffering input if blocked.

        The reasons for the input buffer to block are vaugue
        """
        inp = client.get_input()
        retval = f_send_event(iqueue, 'input', inp)
        # if timeout occured, re-buffer input
        if not retval:
            client.recv_buffer.fromstring(inp)
        return retval

    def session_send():
        """
        Test all sessions for idle timeout, and signal exit to subprocess,
        unregister the session.  Also test for data received by telnet
        client, and send to subprocess as 'input' event.
        """
        # accept session event i/o, such as output
        for sid, tty in terminals():
            # poll about and kick off idle users
            if tty.timeout and tty.client.idle() > tty.timeout:
                send_event = 'exception'
                send_data = Disconnected('timeout: %ds' % (tty.client.idle()))
                tty.iqueue.send((send_event, send_data))
                continue
            elif tty.client.input_ready():
                # input buffered on tcp socket, attempt to send to client
                # with a signal alarm timeout; raising a warning if exceeded.
                if not send_input(tty.client, tty.iqueue):
                    logger.warn('%s input buffer exceeded', sid)
                    tty.client.deactivate()

    def handle_lock(tty, event, data):
        """
        handle locking event of (lock-key, (method, stale))
        """
        method, stale = data
        if method == 'acquire':
            if event in locks:
                # lock already held; check for and display owner, or
                # acquire a lock from a now-deceased session.
                held=False
                for _sid, _tty in terminals():
                    if _sid == locks[event][1] and _sid != tty.sid:
                        logger.debug('[%s] %r not acquired, held by %s.',
                                tty.sid, (event, data), _sid)
                        held=_sid
                        break
                if held is not False:
                    logger.debug('[%s] %r discovered stale lock, previously '
                            'held by %s.', tty.sid, (event, data), held)
                    del locks[event]
            if not event in locks:
                locks[event] = (time.time(), tty.sid)
                tty.iqueue.send((event, True,))
                logger.debug('[%s] %r granted.',
                             tty.sid, (event, data))
            else:
                # caller signals this kind of thread is short-lived, and any
                # existing lock older than 'stale' should be released.
                if (stale is not None
                        and time.time() - locks[event][0] > stale):
                    tty.iqueue.send((event, True,))
                    locks[event] = (time.time(), tty.sid)
                    logger.warn('[%s] %r stale %fs.',
                                tty.sid, (event, data),
                                time.time() - locks[event][0])
                # signal busy with matching event, data=False
                else:
                    tty.iqueue.send((event, False,))
                    logger.debug('[%s] %r not acquired.',
                            tty.sid, (event, data))
        elif method == 'release':
            if not event in locks:
                logger.error('[%s] %r failed: no match',
                             tty.sid, (event, data))
            else:
                del locks[event]
                logger.debug('[%s] %r removed.',
                             tty.sid, (event, data))

    def session_recv(fds):
        """
        receive data waiting for session; all data received from
        subprocess is in form (event, data), and is handled by ipc_recv.

        if stale is not None, elapsed time lock was held to consider stale
        and acquire anyway. no actual locks are held or released, just a
        simple dictionary state/time tracking system.
        """

        disp_handle = lambda handle: ((handle + u' ')
                                      if handle is not None
                                      and 0 != len(handle) else u'')
        disp_origin = lambda client: client.addrport().split(':', 1)[0]

        for sid, tty in terminals():
            while tty.oqueue.fileno() in fds and tty.oqueue.poll():
                # receive data from pipe, unregister if any error,
                try:
                    event, data = tty.oqueue.recv()
                except (EOFError, IOError) as err:
                    # sub-process unexpectedly closed
                    logger.exception(err)
                    unregister(tty)
                    return
                # 'exit' event, unregisters client
                if event == 'exit':
                    unregister(tty)
                    return
                # 'logger' event, propogated upward
                elif event == 'logger':
                    # prefix message with 'ip:port/nick '
                    data.msg = '%s[%s] %s' % (
                        disp_handle(data.handle),
                        disp_origin(tty.client),
                        data.msg)
                    logger.handle(data)
                # 'output' event, buffer for tcp socket
                elif event == 'output':
                    tty.client.send_unicode(ucs=data[0], encoding=data[1])
                # 'remote-disconnect' event, hunt and destroy
                elif event == 'remote-disconnect':
                    send_to = data[0]
                    for _sid, _tty in terminals():
                        if send_to == _sid:
                            send_event = 'exception'
                            send_val = Disconnected(
                                'remote-disconnect by %s' % (sid,))
                            tty.iqueue.send((send_event, send_val))
                            unregister(tty)
                            break
                # 'route': message passing directly from one session to another
                elif event == 'route':
                    logger.debug('route %r', (event, data))
                    tgt_sid, send_event, send_val = data[0], data[1], data[2:]
                    for _sid, _tty in terminals():
                        if tgt_sid == _sid:
                            _tty.iqueue.send((send_event, send_val))
                            break
                # 'global': message broadcasting to all sessions
                elif event == 'global':
                    logger.debug('broadcast %r', (event, data))
                    for _sid, _tty in terminals():
                        if sid != _sid:
                            _tty.iqueue.send((event, data,))
                # 'set-timeout': set user-preferred timeout
                elif event == 'set-timeout':
                    logger.debug('set-timeout %d', data)
                    tty.timeout = data
                # 'db*': access DBProxy API for shared sqlitedict
                elif event.startswith('db'):
                    thread = DBHandler(tty.iqueue, event, data)
                    thread.start()
                # 'lock': access fine-grained bbs-global locking
                elif event.startswith('lock'):
                    handle_lock(tty, event, data)
                else:
                    assert False, 'unhandled %r' % ((event, data),)

    #
    # main event loop
    #
    while True:

        # shutdown, close & delete inactive sockets,
        for fileno, client in [(_fno, _client)
                               for (_fno, _client) in telnetd.clients.items()
                               if not _client.active]:
            logger.info('close %s', telnetd.clients[fileno].addrport(),)
            try:
                client.sock.shutdown(socket.SHUT_RDWR)
            except socket.error:
                pass
            client.sock.close()

            # signal exit to any matching session
            for _sid, tty in terminals():
                if client == tty.client:
                    send_event = 'exception'
                    send_data = Disconnected('deactivated')
                    tty.iqueue.send((send_event, send_data,))
                    break

            # unregister
            del telnetd.clients[fileno]

        # send tcp data, pruning list of any clients d/c during activity
        fds = telnet_send([fileno_telnetd] + telnetd.clients.keys())

        # extend fd list with all session Pipes
        fds.extend([tty.oqueue.fileno() for _sid, tty in terminals()])

        try:
            fds = select.select(fds, [], [], 0.1)[0]
        except select.error as err:
            logger.exception(err)
            fds = list()

        if fileno_telnetd in fds:
            accept()

        # recv telnet data,
        telnet_recv(fds)

        # recv and handle session events,
        session_recv(fds)

        # send any session input data, poll timeout
        session_send()


def timeout_alarm(timeout_time, default):
    """
    Call a function using signal handler, return False if it did not
    return within duration of ``timeout_time``.

    http://pguides.net/python-tutorial/python-timeout-a-function/
    """
    import signal

    class TimeoutException(Exception):
        """ Exception thrown when alarm is caught. """
        pass

    # pylint: disable=W0613
    #         Unused argument 'args'
    def timeout_function(func, *args):
        """ decorator """
        def func2(*args):
            """ function wrapper for decorator """
            def timeout_handler(_signum, _frame):
                """ Raises timeout exception. """
                raise TimeoutException()
            old_handler = signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(timeout_time)
            try:
                retval = func(*args)
            except TimeoutException:
                return default
            finally:
                signal.signal(signal.SIGALRM, old_handler)
            signal.alarm(0)
            return retval
        return func2
    return timeout_function


if __name__ == '__main__':
    exit(main())



# -*- coding: utf-8 -*-
"""
Session engine for x/84, http://github.com/jquast/x84/
"""
import traceback
import logging
import inspect
import struct
import math
import time
import imp
import sys
import os
import io

SESSION = None

def getsession():
    """
    Return session, after a .run() method has been called on any 1 instance.
    """
    return SESSION


def getterminal():
    """
    Return blessings terminal instance of this session.
    """
    return getsession().terminal

def getnode():
    """
    Returns unique session identifier for this session as integer.
    """
    return getsession().node

class Session(object):
    """
    A BBS Session engine. Workflow begins in the ``run()`` method.
    """
    # pylint: disable=R0902,R0904,R0913
    #        Too many instance attributes
    #        Too many public methods
    #        Too many arguments
    TRIM_CP437 = bytes(chr(14) + chr(15)) # HACK
    _encoding = None
    _decoder = None

    def __init__(self, terminal, inp_queue, out_queue,
                 sid, env, lock, encoding='utf8'):
        """
        Instantiate a Session instanance, only one session
        may be instantiated per process. Arguments:
            terminal: blessings.Terminal,
            inp_queue: multiprocessing.Queue Parent writes, Child reads
            out_queue: multiprocessing.Queue Parent reads, Child writes
            sid: session id by engine: origin of telnet connection (ip:port),
            env: dict of environment variables, such as 'TERM', 'USER'.
        """
        from x84.bbs import ini
        # pylint: disable=W0603
        #        Using the global statement
        global SESSION
        assert SESSION is None, 'Session may be instantiated only once'
        SESSION = self
        self.iqueue = inp_queue
        self.oqueue = out_queue
        self.terminal = terminal
        self.sid = sid
        self.env = env
        self.encoding = encoding
        self.lock = lock
        self._user = None
        self._script_stack = [(ini.CFG.get('matrix', 'script'),)]
        self._tap_input = ini.CFG.getboolean('session', 'tap_input')
        self._tap_output = ini.CFG.getboolean('session', 'tap_output')
        self._tap_events = ini.CFG.getboolean('session', 'tap_events')
        self._ttyrec_folder = ini.CFG.get('system', 'ttyrecpath')
        self._record_tty = ini.CFG.getboolean('session', 'record_tty')
        self._show_traceback = ini.CFG.getboolean('system', 'show_traceback')
        self._script_path = ini.CFG.get('system', 'scriptpath')
        self._script_module = None
        self._fp_ttyrec = None
        self._ttyrec_fname = None
        self._node = None
        self._connect_time = time.time()
        self._last_input_time = time.time()
        self._activity = u'<uninitialized>'
        # event buffer
        self._buffer = dict()
        # save state for ttyrec compression
        self._ttyrec_sec = -1
        self._ttyrec_usec = -1
        self._ttyrec_len_text = 0

    @property
    def duration(self):
        """
        Return length of time since connection began (float).
        """
        return time.time() - self._connect_time

    @property
    def connect_time(self):
        """
        Return time when connection began (float).
        """
        return self._connect_time

    @property
    def last_input_time(self):
        """
        Return last time of keypress (epoch float).
        """
        return self._last_input_time

    @property
    def idle(self):
        """
        Return length of time since last keypress occured (float).
        """
        return time.time() - self._last_input_time

    @property
    def activity(self):
        """
        Current activity (arbitrarily set). This also updates xterm titles,
        and is globally broadcasted as a "current activity" in the Who's
        online script.
        """
        return self._activity

    @activity.setter
    def activity(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        if self._activity != value:
            logger = logging.getLogger()
            logger.debug('activity=%s', value)
            term = self.env.get('TERM', 'unknown')
            set_title = self.user.get('set-title', (
                'xterm' in term or 'rxvt' in term
                or '_xtitle' in self.env))
            self._activity = value
            if set_title:
                self.write(u''.join((
                    unichr(27), u']2;%s' % (value,), unichr(7))))

    @property
    def handle(self):
        """
        Returns User handle.
        """
        return self.user.handle

    @property
    def user(self):
        """
        User record of session.
        """
        from x84.bbs.userbase import User
        if self._user is not None:
            return self._user
        return User()

    @user.setter
    def user(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._user = value
        logger = logging.getLogger()
        logger.info("set user '%s'.", value.handle)

    @property
    def encoding(self):
        """
        Session terminal encoding; only 'utf8' and 'cp437' are supported.
        """
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        import codecs
        if value != self._encoding:
            logger = logging.getLogger()
            logger.info('encoding is %s.', value)
            assert value in ('utf8', 'cp437')
            self._encoding = value
            self._decoder = codecs.getincrementaldecoder(value)()

    @property
    def pid(self):
        """
        Returns Process ID.
        """
        # pylint: disable=R0201
        #        Method could be a function
        return os.getpid()

    @property
    def node(self):
        """
        Returns numeric constant for session, often required by 'doors'
        """
        if self._node is None:
            for node in range(1, 64):
                event = 'lock-%s/%d' % ('node', node)
                self.send_event(event, ('acquire', None))
                data = self.read_event(event)
                if data is True:
                    self._node = node
                    break
        return self._node

    def __error_recovery(self):
        """
        jojo's invention; recover from a general exception by using
        a script stack, and resuming last good script.
        """
        logger = logging.getLogger()
        if 0 != len(self._script_stack):
            # recover from exception
            fault = self._script_stack.pop()
            oper = 'RESUME' if len(self._script_stack) else 'STOP'
            stop = bool(0 == len(self._script_stack))
            msg = (u'%s %safter general exception in %s.' % (
                oper, (
                    (self._script_stack[-1][0] + u' ')
                    if len(self._script_stack) else u' '),
                fault[0],))
            logger.info(msg)
            self.write(u'\r\n\r\n')
            if stop:
                self.write(self.terminal.red_reverse('stop'))
            else:
                self.write(self.terminal.bold_green('continue'))
                self.write(u' ' + self.terminal.bold_cyan(
                    self._script_stack[-1][0]))
            self.write(u' after general exception in %s\r\n' % (
                self.terminal.bold_cyan(fault[0]),))
            # give time for exception to write down queue before
            # continuing or exiting, esp. exiting, otherwise
            # STOP message is not often fully received
            time.sleep(2)

    def run(self):
        """
        Begin main execution flow.

        Scripts manipulate control flow of scripts using goto and gosub.
        """
        from x84.bbs.exception import Goto, Disconnected
        logger = logging.getLogger()
        while len(self._script_stack):
            logger.debug('script_stack: %r', self._script_stack)
            try:
                self.runscript(*self._script_stack.pop())
                continue
            except Goto, err:
                logger.debug('Goto: %s', err)
                self._script_stack = [err[0] + tuple(err[1:])]
                continue
            except Disconnected, err:
                logger.info('Disconnected: %s', err)
                self.close()
                return None
            except Exception, err:
                # Pokemon exception, log and Cc: telnet client, then resume.
                e_type, e_value, e_tb = sys.exc_info()
                if self._show_traceback:
                    self.write(self.terminal.normal + u'\r\n')
                terrs = list()
                for line in traceback.format_tb(e_tb):
                    for subln in line.split('\n'):
                        terrs.append(subln)
                terrs.extend(traceback.format_exception_only(e_type, e_value))
                for etxt in terrs:
                    logger.error(etxt.rstrip())
                    if self._show_traceback:
                        self.write(etxt.rstrip() + u'\r\n')
            self.__error_recovery()
        logger.info('End of script stack.')
        self.close()
        return None

    def write(self, ucs):
        """
        Write unicode data to telnet client. Take special care to encode
        as 'iso8859-1' actually intended for 'cp437'-encoded terminals.

        Has side effect of updating ttyrec file when recording.
        """
        from x84.bbs.cp437 import CP437
        logger = logging.getLogger()
        if 0 == len(ucs):
            return
        assert isinstance(ucs, unicode)
        if self.encoding == 'cp437':
            encoding = 'iso8859-1'
            # out output terminal is cp437, so we need to take special care to
            # re-encode things as "iso8859-1" but really encoded for cp437.
            # For example, u'\u2591' becomes u'\xb0' (unichr(176)),
            # -- the original ansi shaded block for cp437 terminals.
            #
            # additionally, the 'shift-in' and 'shift-out' characters
            # display as '*' on SyncTerm, I think they stem from curses:
            # http://lkml.indiana.edu/hypermail/linux/kernel/0602.2/0868.html
            # regardless, remove them using str.translate()
            text = ucs.encode(encoding, 'replace')
            ucs = u''.join([(unichr(CP437.index(glyph))
                             if glyph in CP437
                             and not glyph in self.TRIM_CP437
                             else unicode(
                                 text[idx].translate(None, self.TRIM_CP437),
                                 encoding, 'replace'))
                            for (idx, glyph) in enumerate(ucs)])
        else:
            encoding = self.encoding
        self.terminal.stream.write(ucs, encoding)

        if self._tap_output and logger.isEnabledFor(logging.DEBUG):
            logger.debug('--> %r.', ucs)

        if self._record_tty:
            if not self.is_recording:
                self.start_recording()
            self._ttyrec_write(ucs)

    def flush_event(self, event):
        """
        Flush all return all data buffered for 'event'.
        """
        logger = logging.getLogger()
        flushed = list()
        while True:
            data = self.read_event(event, -1)
            if data is None:
                if 0 != len(flushed):
                    logger.debug('flushed from %s: %r', event, flushed)
                return flushed
            flushed.append(data)
        return flushed

    def info(self):
        """
        Returns dictionary of key, value pairs of session paramters.
        """
        return dict((
            ('TERM', self.env.get('TERM', u'unknown')),
            ('LINES', self.terminal.height,),
            ('COLUMNS', self.terminal.width,),
            ('sid', self.sid,),
            ('handle', self.user.handle,),
            ('script', (self._script_stack[-1][0]
                        if len(self._script_stack) else None)),
            ('ttyrec',
             self._fp_ttyrec.name if self._fp_ttyrec is not None else u'',),
            ('connect_time', self.connect_time),
            ('idle', self.idle),
            ('activity', self.activity),
            ('encoding', self.encoding),
            ('node', self._node),
        ))

    def buffer_event(self, event, data=None):
        """
        Push data into buffer keyed by event. Handle special events:
            'exception', 'global' AYT (are you there),
            'page', 'info-req', 'refresh', and 'input'.
        """
        logger = logging.getLogger()
        # exceptions aren't buffered; they are thrown!
        if event == 'exception':
            # pylint: disable=E0702
            #        Raising NoneType while only classes, (..) allowed
            raise data

        # respond to global 'AYT' requests
        if event == 'global' and data[0] == 'AYT':
            reply_to = data[1]
            self.send_event('route', (
                reply_to, 'ACK',
                self.sid, self.user.handle,))
            return True

        # accept 'page' as instant chat when 'mesg' is True, or sender is -1
        # -- intent is that sysop can always 'chat' a user ..
        if event == 'page' and self._script_stack[-1:][0][0] != 'chat':
            channel, sender = data
            if self.user.get('mesg', True) or sender == -1:
                logger.info('page from %s.' % (sender,))
                if not self.runscript('chat', channel, sender):
                    logger.info('rejected page from %s.' % (sender,))
                # buffer refresh event for any asyncronous event UI's
                self.buffer_event('refresh', 'page-return')
                return True

        # respond to 'info-req' events by returning pickled session info
        if event == 'info-req':
            sid = data[0]
            self.send_event('route', (sid, 'info-ack', self.sid, self.info(),))
            return True

        # init new unmanaged & unlimited-sized buffer ;p
        if not event in self._buffer:
            self._buffer[event] = list()

        # buffer input
        if event == 'input':
            self.buffer_input(data)
            return

        # buffer only 1 most recent 'refresh' event
        if event == 'refresh':
            if data[0] == 'resize':
                # inherit terminal dimensions values
                (self.terminal.columns, self.terminal.rows) = data[1]
            # store only most recent 'refresh' event
            self._buffer[event] = list((data,))
            return True

        # buffer all else
        self._buffer[event].insert(0, data)

        # global events are meant to be missed if unwanted, so
        # we keep only the 100 most recent.
        if event == 'global' and len(self._buffer[event]) > 150:
            self._buffer[event] = self._buffer[event][:100]

    def buffer_input(self, data):
        """
        Update idle time, buffering raw bytes received from telnet client
        via event queue
        """
        self._last_input_time = time.time()
        ctrl_l = self.terminal.KEY_REFRESH

        logger = logging.getLogger()
        if self._tap_input and logger.isEnabledFor(logging.DEBUG):
            logger.debug('<-- (%d): %r.', len(data), data)

        for keystroke in data:
            self._buffer['input'].insert(0, keystroke)
        return

    def send_event(self, event, data):
        """
           Send data to IPC output queue in form of (event, data).

           Supported events:
               'disconnect': Session wishes to disconnect.
               'logger': Data is logging record, used by IPCLogHandler.
               'output': Unicode data to write to client.
               'global': Broadcast event to other sessions.
               XX 'pos': Request cursor position.
               'db-<schema>': Request sqlite dict method result.
               'db=<schema>': Request sqlite dict method result as iterable.
               'lock-<name>': Fine-grained global bbs locking.
        """
        self.lock.acquire()
        self.oqueue.send((event, data))
        self.lock.release()

    def poll_event(self, event):
        """
        Non-blocking poll for session event, returns value, if any. None
        otherwise.
        """
        return self.read_event(event, -1)

    def read_event(self, event, timeout=None):
        """
        S.read_event (event, timeout=None) --> data

        Read any data for a single event.

        Blocking by default, or non-blocking when timeout is -1. When timeout
        is non-zero, specifies length of time to wait for event before
        returning. If timeout is not None (non-blocking), None is returned if
        no event has is waiting, or waiting after timeout has elapsed.
        """
        return self.read_events(events=(event,), timeout=timeout)[1]

    def read_events(self, events, timeout=None):
        """
           S.read_events (events, timeout=None) --> (event, data)

           Return the first matched IPC data for any event specified in tuple
           events, in the form of (event, data).
        """
        logger = logging.getLogger()
        (event, data) = (None, None)
        # return immediately any events that are already buffered
        for (event, data) in ((e, self._event_pop(e))
                              for e in events if e in self._buffer
                              and 0 != len(self._buffer[e])):
            return (event, data)
        stime = time.time()
        timeleft = lambda cmp_time: (
            float('inf') if timeout is None else
            timeout if timeout < 0 else
            timeout - (time.time() - cmp_time))
        waitfor = timeleft(stime)
        while waitfor > 0:
            poll = None if waitfor == float('inf') else waitfor
            if self.iqueue.poll(poll):
                event, data = self.iqueue.recv()
                retval = self.buffer_event(event, data)
                if (self._tap_events and logger.isEnabledFor(logging.DEBUG)):
                    stack = inspect.stack()
                    caller_mod, caller_func = stack[2][1], stack[2][3]
                    logger.debug('event %s %s by %s in %s.', event,
                                 'caught' if event in events else
                                 'handled' if retval is not None else
                                 'buffered', caller_func, caller_mod,)
                if event in events:
                    return (event, self._event_pop(event))
            elif timeout == -1:
                return (None, None)
            waitfor = timeleft(stime)
        return (None, None)

    def _event_pop(self, event):
        """
        S._event_pop (event) --> data

        Returns foremost item buffered for event. When event is ``input``,
        an artificial pause is used for decoding of MBS when received multipart
        """
        return self._buffer[event].pop()

    def runscript(self, script_name, *args):
        """
        Execute the main() callable of script identified by
        *script_name*, with optional args.
        """
        from x84.bbs.exception import ScriptError
        logger = logging.getLogger()
        self._script_stack.append((script_name,) + args)
        logger.info("run script '%s'%s.", script_name,
                    ', args %r' % (args,) if 0 != len(args) else '')

        def _load_script_module():
            """
            Load and return ini folder, `scriptpath` as a module (cached).
            """
            if self._script_module is None:
                # load default/__init__.py as 'default',
                base_script = os.path.basename(self._script_path)
                lookup = imp.find_module(script_name, [self._script_path])
                # pylint: disable=W0142
                #        Used * or ** magic
                self._script_module = imp.load_module(base_script, *lookup)
                self._script_module.__path__ = self._script_path
            return self._script_module
        # pylint: disable=W0142
        #        Used * or ** magic
        script_module = _load_script_module()
        lookup = imp.find_module(script_name, [script_module.__path__])
        script = imp.load_module(script_name, *lookup)
        if not hasattr(script, 'main'):
            raise ScriptError("%s: main() not found." % (script_name,))
        if not callable(script.main):
            raise ScriptError("%s: main not callable." % (script_name,))
        value = script.main(*args)
        toss = self._script_stack.pop()
        logger.info("script '%s' returned%s.", toss[0],
                    ' %r' % (value,) if value is not None else u'')
        return value

    def close(self):
        """
        Close session.
        """
        if self.is_recording:
            self.stop_recording()
        if self._node is not None:
            self.send_event(
                    event='lock-node/%d' % (self._node),
                    data=('release', None))

    @property
    def is_recording(self):
        """
        True when session is being recorded to ttyrec file
        """
        return self._fp_ttyrec is not None

    def stop_recording(self):
        """
        Cease recording to ttyrec file (close).
        """
        assert self.is_recording
        self._ttyrec_write(u''.join((
            self.terminal.normal, u'\r\n\r\n',
            u'\r\n'.join([u'%s: %s' % (key, val)
            for (key, val) in sorted(self.info().items())]),
            u'\r\n')))
        self._fp_ttyrec.close()
        self._fp_ttyrec = None

    def start_recording(self):
        """
        Begin recording to ttyrec file.
        """
        logger = logging.getLogger()
        assert self._fp_ttyrec is None, ('already recording')
        digit = 0
        while True:
            self._ttyrec_fname = '%s%d-%s.ttyrec' % (
                time.strftime('%Y%m%d.%H%M%S'), digit,
                self.sid.split(':', 1)[0],)
            if not os.path.exists(self._ttyrec_fname):
                break
            digit += 1
        assert os.path.sep not in self._ttyrec_fname
        filename = os.path.join(self._ttyrec_folder, self._ttyrec_fname)
        if not os.path.exists(self._ttyrec_folder):
            logger.info('creating ttyrec folder, %s.', self._ttyrec_folder)
            os.makedirs(self._ttyrec_folder)
        self._fp_ttyrec = io.open(filename, 'wb+')
        self._ttyrec_sec = -1
        self._ttyrec_write_header()
        logger.info('recording %s.' % (filename,))

    def _ttyrec_write_header(self):
        """
        Write ttyrec header that identifies termianl height & width, and escape
        sequence to indicate UTF-8 mode.
        """
        (height, width) = self.terminal.height, self.terminal.width
        self._ttyrec_write(unichr(27) + u'[8;%d;%dt' % (height, width,))
        # ESC %G activates UTF-8 with an unspecified implementation level from
        # ISO 2022 in a way that allows to go back to ISO 2022 again.
        self._ttyrec_write(unichr(27) + u'%G')

    def _ttyrec_write(self, ucs):
        """
        Update ttyrec stream with unicode bytes 'ucs'.
        """
        # write bytestring to ttyrec file packed as timed byte.
        # If the current timed byte is within TTYREC_UCOMPRESS
        # (default: 15,000 μsec), rewind stream and re-write the
        # 'length' portion, and append data to end of stream.
        # .. unfortuantely, this is not compatible with ttyplay -p,
        # so for the time being, it is disabled ..
        assert self._fp_ttyrec is not None, 'call start_recording() first'
        timekey = self.duration

        # Round down timekey to nearest whole number,
        # use the remainder for microseconds. Upconvert,
        # constructing a (seconds, microseconds) pair.
        sec = math.floor(timekey)
        usec = (timekey - sec) * 1e+6
        sec, usec = int(sec), int(usec)

        def write_chunk(tm_sec, tm_usec, textlen, u_text):
            """
            Write new timechunk record,
              bytes (sec, usec, len(text), text.. )
            """
            # build & write,
            bp1 = struct.pack('<I', tm_sec) + struct.pack('<I', tm_usec)
            bp2 = struct.pack('<I', textlen)
            self._fp_ttyrec.write(bp1 + bp2 + u_text)
            # save (time,len) state for compression
            self._ttyrec_sec = tm_sec
            self._ttyrec_usec = tm_usec
            self._ttyrec_len_text = textlen
            self._fp_ttyrec.flush()
        text = ucs.encode('utf8', 'replace')
        len_text = len(text)
        # TODO: padd every 1 or 10 seconds, so 'VCR' type apps
        # can FF/RW easier
        return write_chunk(sec, usec, len_text, text)



from contextlib import contextmanager
import curses
import curses.has_key
import curses.ascii
from curses import tigetstr, tigetnum, setupterm, tparm
try:
    from io import UnsupportedOperation as IOUnsupportedOperation
except ImportError:
    class IOUnsupportedOperation(Exception):
        """A dummy exception to take the place of Python 3's ``io.UnsupportedOperation`` in Python 2"""
import os
from platform import python_version_tuple
import textwrap
import codecs
import locale
import struct
import time
import math
import sys
import re

import termios
import select
import fcntl
import tty

__all__ = ['Terminal']


if ('3', '0', '0') <= python_version_tuple() < ('3', '2', '2+'):  # Good till 3.2.10
    # Python 3.x < 3.2.3 has a bug in which tparm() erroneously takes a string.
    raise ImportError('Blessings needs Python 3.2.3 or greater for Python 3 '
                      'support due to http://bugs.python.org/issue10570.')

MMODE_X10 = 9
MMODE_VT200 = 1000
MMODE_VT200_HIGHLIGHT = 1001
MMODE_BTN_EVENT = 1002
MMODE_ANY_EVENT = 1003
MMODE_FOCUS_EVENT = 1004
MMODE_EXT_MODE = 1005
MMODE_SGR_EXT_MODE = 1006
MMODE_URXVT_EXT_MODE = 1015
MMODE_ALTERNATE_SCROLL = 1007
MMODE_XTERM_262 = 262
MMODE_URXVT_910 = 910
MMODE_DEFAULT = MMODE_URXVT_910

class Terminal(object):
    """An abstraction around terminal capabilities

    Unlike curses, this doesn't require clearing the screen before doing
    anything, and it's friendlier to use. It keeps the endless calls to
    ``tigetstr()`` and ``tparm()`` out of your code, and it acts intelligently
    when somebody pipes your output to a non-terminal.

    Instance attributes:

      ``stream``
        The stream the terminal outputs to. It's convenient to pass the stream
        around with the terminal; it's almost always needed when the terminal
        is and saves sticking lots of extra args on client functions in
        practice.
      ``encoding``
        The encoding used for keyboard input in the ``inkey()`` method; by
        default it is the preferred encoding of the environment locale.
    """
    def __init__(self, kind=None, stream=None, force_styling=False):
        """Initialize the terminal.

        If ``stream`` is not a tty, I will default to returning an empty
        Unicode string for all capability values, so things like piping your
        output to a file won't strew escape sequences all over the place. The
        ``ls`` command sets a precedent for this: it defaults to columnar
        output when being sent to a tty and one-item-per-line when not.

        :arg kind: A terminal string as taken by ``setupterm()``. Defaults to
            the value of the ``TERM`` environment variable.
        :arg stream: A file-like object representing the terminal. Defaults to
            the original value of stdout, like ``curses.initscr()`` does.
        :arg force_styling: Whether to force the emission of capabilities, even
            if we don't seem to be in a terminal. This comes in handy if users
            are trying to pipe your output through something like ``less -r``,
            which supports terminal codes just fine but doesn't appear itself
            to be a terminal. Just expose a command-line option, and set
            ``force_styling`` based on it. Terminal initialization sequences
            will be sent to ``stream`` if it has a file descriptor and to
            ``sys.__stdout__`` otherwise. (``setupterm()`` demands to send them
            somewhere, and stdout is probably where the output is ultimately
            headed. If not, stderr is probably bound to the same terminal.)

            If you want to force styling to not happen, pass
            ``force_styling=None``.

        """
        # any desire for passing in a 'virtual keyboard' input stream??
        i_stream = sys.__stdin__
        if stream is None:
            o_stream = sys.__stdout__
        else:
            o_stream = stream
        # for StringIO's and such, a file descriptor is not necessarily
        # always available for the output stream; that is ok, though, as
        # long as the stream offers a .write method.
        try:
            o_fd = (o_stream.fileno() if hasattr(o_stream, 'fileno')
                    and callable(o_stream.fileno) else None)
        except IOUnsupportedOperation:
            o_fd = None

        # os.isatty returns True if output stream is an open file
        # descriptor connected to the slave end of a terminal.
        self._is_a_tty = o_fd is not None and os.isatty(o_fd)
        self._do_styling = ((self.is_a_tty or force_styling) and
                            force_styling is not None)

        # an open file descriptor of some sort is necessary for setupterm(),
        # even though it goes unused when an alternative stream is specified.
        self.o_fd = sys.__stdout__.fileno() if o_fd is None else o_fd

        if self.do_styling:
            # curses.setupterm is the foot in the water, providing access to
            # the extensive terminal capabilities database so necessary for
            # blessings.
            # Note: we go through great lengths to prevent diving into curses
            # fully by avoiding a call to curses.initscr().
            # Prefer the stream file descriptor if available, using stdout as a
            # a fallback. Even though it is unused, it is required by setuperm.
            setupterm(kind or os.environ.get('TERM', 'unknown'), self.o_fd)

        self.stream = o_stream
        self.i_stream = i_stream

        # a beginning state of echo ON and canonical mode is assumed.
        self._state_echo = True
        self._state_canonical = True

        # dictionary of multibyte sequences to be paired with key codes
        self._keymap = {}
        # list of key code names
        self._keycodes = []

        # Inherit curses keycap capability names, such as KEY_DOWN, to be
        # used with Keystroke ``code`` property values for comparison to the
        # Terminal class instance it was received on.
        max_int = 256 # curses keycode values begin beyond 8-bit range
        for keycode in [kc for kc in dir(curses) if kc.startswith('KEY_')]:
            value = getattr(curses, keycode)
            self._keycodes.append(keycode)
            setattr(self, keycode, value)
            max_int = max(max_int, value)
        # Holy smokes; I made this up! Ctrl+Space sends \x00.
        # Ctrl+Space is used in interfaces fe. midnight commander.
        self.KEY_CSPACE = max_int + 1

        if self.is_a_tty:
            # determine encoding of input stream. Only used for keyboard
            # input on posix systems. win32 systems use getwche which returns
            # unicode, and does not require decoding.
            locale.setlocale(locale.LC_ALL, '')
            self.encoding = locale.getpreferredencoding()
            if sys.platform != 'win32':
                self._idecoder = codecs.getincrementaldecoder(self.encoding)()
            self.i_buf = []

            # create lookup dictionary for multibyte keyboard input sequences
            self._init_keystrokes()

        # Friendly mnemonics for 'KEY_DELETE', 'KEY_INSERT', 'KEY_PGUP',
        # and 'KEY_PGDOWN', 'KEY_SPGUP', 'KEY_SPGDOWN'.
        self.KEY_DELETE = self.KEY_DC
        self.KEY_INSERT = self.KEY_IC
        self.KEY_PGUP = self.KEY_PPAGE
        self.KEY_PGDOWN = self.KEY_NPAGE
        self.KEY_SUP = self.KEY_SR  # scroll reverse (shift+pgup)
        self.KEY_SDOWN = self.KEY_SF  # scroll forward (shift+pgdown)
        self.KEY_ESCAPE = self.KEY_EXIT

    # Sugary names for commonly-used capabilities, intended to help avoid trips
    # to the terminfo man page and comments in your code:
    _sugar = dict(
        # Don't use "on" or "bright" as an underscore-separated chunk in any of
        # these (e.g. on_cology or rock_on) so we don't interfere with
        # __getattr__.
        save='sc',
        restore='rc',

        clear_eol='el',
        clear_bol='el1',
        clear_eos='ed',
        # 'clear' clears the whole screen.
        position='cup',  # deprecated
        enter_fullscreen='smcup',
        exit_fullscreen='rmcup',
        move='cup',
        move_x='hpa',
        move_y='vpa',
        move_left='cub1',
        move_right='cuf1',
        move_up='cuu1',
        move_down='cud1',

        hide_cursor='civis',
        normal_cursor='cnorm',

        reset_colors='op',  # oc doesn't work on my OS X terminal.

        normal='sgr0',
        reverse='rev',
        # 'bold' is just 'bold'. Similarly...
        # blink
        # dim
        # flash
        italic='sitm',
        no_italic='ritm',
        shadow='sshm',
        no_shadow='rshm',
        standout='smso',
        no_standout='rmso',
        subscript='ssubm',
        no_subscript='rsubm',
        superscript='ssupm',
        no_superscript='rsupm',
        underline='smul',
        no_underline='rmul')

    def _init_keystrokes(self):
        # curses.has_key._capability_names is a dictionary keyed by termcap
        # capabilities, with integer values to be paired with KEY_ names;
        # using this dictionary, we query for the terminal sequence of the
        # terminal capability, and, if any result is found, store the sequence
        # in the _keymap lookup table with the integer valued to be paired by
        # key codes.
        for i_val, capability in curses.has_key._capability_names.iteritems():
            seq = curses.tigetstr(capability)
            if seq is not None:
                self._keymap[seq.decode('iso8859-1')] = i_val

        # monkey-patch non-destructive space as KEY_RIGHT,
        # in 'xterm-256color' 'kcuf1' is '\x1bOC' and 'cuf1' = '\x1b[C'.
        # xterm sends '\x1b[C'
        ndsp = curses.tigetstr('cuf1')
        if ndsp is not None:
            self._keymap[ndsp.decode('iso8859-1')] = self.KEY_RIGHT

        # ... as well as a list of general NVT sequences you would
        # expect to receive from any remote terminals. Notably the
        # variables of ENTER (^J and ^M), backspace (^H), delete (127),
        # and common sequencess across putty, rxvt, kermit, minicom,
        # SyncTerm, windows telnet, HyperTerminal, netrunner ...
        self._keymap.update([
            (unichr(10), self.KEY_ENTER), (unichr(13), self.KEY_ENTER),
            (unichr(8), self.KEY_BACKSPACE),
            (u"\x1bOA", self.KEY_UP),    (u"\x1bOB", self.KEY_DOWN),
            (u"\x1bOC", self.KEY_RIGHT), (u"\x1bOD", self.KEY_LEFT),
            (u"\x1bOH", self.KEY_LEFT),
            (u"\x1bOF", self.KEY_END),
            (u"\x1b[A", self.KEY_UP),    (u"\x1b[B", self.KEY_DOWN),
            (u"\x1b[C", self.KEY_RIGHT), (u"\x1b[D", self.KEY_LEFT),
            (u"\x1b[U", self.KEY_NPAGE), (u"\x1b[V", self.KEY_PPAGE),
            (u"\x1b[H", self.KEY_HOME),  (u"\x1b[F", self.KEY_END),
            (u"\x1b[K", self.KEY_END),
            (u"\x1bA", self.KEY_UP),     (u"\x1bB", self.KEY_DOWN),
            (u"\x1bC", self.KEY_RIGHT),  (u"\x1bD", self.KEY_LEFT),
            (u"\x1b?x", self.KEY_UP),    (u"\x1b?r", self.KEY_DOWN),
            (u"\x1b?v", self.KEY_RIGHT), (u"\x1b?t", self.KEY_LEFT),
            (u"\x1b[@", self.KEY_IC),  # insert
            (unichr(127), self.KEY_BACKSPACE),  # ^? is backspace (int 127)
        ])

        # windows 'multibyte' translation, not tested.
        if sys.platform == 'win32':
            # http://msdn.microsoft.com/en-us/library/aa299374%28VS.60%29.aspx
            self._keymap.update([
                (u'\xe0\x48', self.KEY_UP),    (u'\xe0\x50', self.KEY_DOWN),
                (u'\xe0\x4D', self.KEY_RIGHT), (u'\xe0\x4B', self.KEY_LEFT),
                (u'\xe0\x51', self.KEY_NPAGE), (u'\xe0\x49', self.KEY_PPAGE),
                (u'\xe0\x47', self.KEY_HOME),  (u'\xe0\x4F', self.KEY_END),
                (u'\xe0\x3B', self.KEY_F1),    (u'\xe0\x3C', self.KEY_F2),
                (u'\xe0\x3D', self.KEY_F3),    (u'\xe0\x3E', self.KEY_F4),
                (u'\xe0\x3F', self.KEY_F5),    (u'\xe0\x40', self.KEY_F6),
                (u'\xe0\x41', self.KEY_F7),    (u'\xe0\x42', self.KEY_F8),
                (u'\xe0\x43', self.KEY_F9),    (u'\xe0\x44', self.KEY_F10),
                (u'\xe0\x85', self.KEY_F11),   (u'\xe0\x86', self.KEY_F12),
                (u'\xe0\x4C', self.KEY_B2),  # center
                (u'\xe0\x52', self.KEY_IC),  # insert
                (u'\xe0\x53', self.KEY_DC),  # delete
            ])

    def __getattr__(self, attr):
        """Return parametrized terminal capabilities, like bold.

        For example, you can say ``term.bold`` to get the string that turns on
        bold formatting and ``term.normal`` to get the string that turns it off
        again. Or you can take a shortcut: ``term.bold('hi')`` bolds its
        argument and sets everything to normal afterward. You can even combine
        things: ``term.bold_underline_red_on_bright_green('yowzers!')``.

        For a parametrized capability like ``cup``, pass the parameters too:
        ``some_term.cup(line, column)``.

        ``man terminfo`` for a complete list of capabilities.

        Return values are always Unicode.

        """
        resolution = (self._resolve_formatter(attr) if self.do_styling
                      else NullCallableString())
        setattr(self, attr, resolution)  # Cache capability codes.
        return resolution

    @property
    def do_styling(self):
        """Wether the terminal will attempt to output sequences for
        all styling attributes, such as term.bold(). When False, a
        null string ('') is used for all styling attributes, this behavior
        is forced when force_styling=None, or when the output is not a tty.
           """
        return self._do_styling

    @property
    def is_a_tty(self):
        """Wether the stream is connected to the slave end of a Terminal.
        This property is used to identify if the terminal can be interacted
        with, for instance, with the ``inkey()`` method.  """
        return self._is_a_tty

    @property
    def height(self):
        """The height of the terminal in characters

        If no stream or a stream not representing a terminal was passed in at
        construction, return the dimension of the controlling terminal so
        piping to things that eventually display on the terminal (like ``less
        -R``) work. If a stream representing a terminal was passed in, return
        the dimensions of that terminal. If there somehow is no controlling
        terminal, return ``None``. (Thus, you should check that ``is_a_tty`` is
        true before doing any math on the result.)

        """
        return self._height_and_width()[0]

    @property
    def width(self):
        """The width of the terminal in characters

        See ``height()`` for some corner cases.

        """
        return self._height_and_width()[1]

    def _height_and_width_win32(self):
        # based on anatoly techtonik's work from pager.py, MIT/Pub Domain.
        # completely untested ... please report !!
        # WIN32_I_FD = -10
        WIN32_O_FD = -11
        from ctypes import windll, Structure, byref
        from ctypes.wintypes import SHORT, WORD, DWORD
        console_handle = windll.kernel32.GetStdHandle(WIN32_O_FD)
            # CONSOLE_SCREEN_BUFFER_INFO Structure

        class COORD(Structure):
            _fields_ = [("X", SHORT), ("Y", SHORT)]

        class SMALL_RECT(Structure):
            _fields_ = [("Left", SHORT), ("Top", SHORT),
                        ("Right", SHORT), ("Bottom", SHORT)]

        class CONSOLE_SCREEN_BUFFER_INFO(Structure):
            _fields_ = [("dwSize", COORD),
                        ("dwCursorPosition", COORD),
                        ("wAttributes", WORD),
                        ("srWindow", SMALL_RECT),
                        ("dwMaximumWindowSize", DWORD)]
        sbi = CONSOLE_SCREEN_BUFFER_INFO()
        ret = windll.kernel32.GetConsoleScreenBufferInfo(
            console_handle, byref(sbi))
        if ret != 0:
            return (sbi.srWindow.Right - sbi.srWindow.Left + 1,
                    sbi.srWindow.Bottom - sbi.srWindow.Top + 1)
        return None, None

    def _height_and_width(self):
        """Return a tuple of (terminal height, terminal width).
           Returns (None, None) if terminal window size is indeterminate.
       """
        # tigetnum('lines') and tigetnum('cols') update only if we call
        # setupterm() again.
        buf = struct.pack('HHHH', 0, 0, 0, 0)
        if sys.platform == 'win32':
            value = self._height_and_width_win32()
            if value is not None:
                return value
        for fd in self.o_fd, sys.__stdout__.fileno():
            try:
                value = fcntl.ioctl(fd, termios.TIOCGWINSZ, buf)
                return struct.unpack('hhhh', value)[0:2]
            except IOError:
                pass
        # throw a hail mairy for environment values
        lines, cols = (os.environ.get('LINES', None),
                       os.environ.get('COLUMNS', None))
        if None not in (lines, cols):
            return lines, cols
        # ! should never be reached
        return None, None

    def kbhit(self, timeout=0):
        """ Returns True if a keypress has been detected on input.
        A subsequent call to getch() will not block on cbreak mode.
        A timeout of 0 returns immediately (default), A timeout of
        ``None`` blocks indefinitely. A timeout of non-zero blocks
        until timeout seconds elapsed. """
        fds = select.select([self.i_stream.fileno()], [], [], timeout)[0]
        return self.i_stream.fileno() in fds

    def getch(self):
        """ Read a single byte from input stream. """
        return os.read(self.i_stream.fileno(), 1)

    def kbflush(self):
        """ Flush buffered keystrokes, returns number of keystrokes removed """
        n_removed = 0
        for n_removed in range(len(self.i_buf)):
            self.i_buf.pop()
        return n_removed

    def inkey(self, timeout=None, esc_delay=0.35):
        """ Read a single keystroke from input up to timeout in seconds,
        translating special application keys, such as KEY_LEFT.

        Ensure you use 'cbreak mode', by using the ``cbreak`` context
        manager, otherwise input is not received until return is pressed!

        When ``timeout`` is None (default), block until input is available.
        If ``timeout`` is 0, this function is non-blocking and None is
        returned if no input is available.  If ``timeout`` is non-zero,
        None is returned after ``timeout`` seconds have elapsed without input.

        This method differs from using kbhit in combination with getch in that
        Multibyte sequences are translated to key code values, and string
        sequence of length greater than 1 may be returned.

        The result is a unicode-typed instance of the Keystroke class, with
        additional properties ``is_sequence`` (bool), ``name`` (str),
        and ``value`` (int). esc_delay defines the time after which
        an escape key is pressed that the stream awaits a multibyte sequence.

        with term.cbreak():
            inp = None
            while inp not in (u'q', u'Q'):
                inp = term.inkey(3)
                if inp is None:
                    print 'timeout after 3 seconds'
                elif inp.code == term.KEY_UP:
                        print 'moving up!'
                else:
                    print 'pressed', 'sequence' if inp.is_sequence else 'ascii'
                    print 'key:', inp, inp.code
        """
        assert self.is_a_tty, u'stream is not a a tty.'
        esc = curses.ascii.ESC  # posix multibyte sequence (MBS) start mark
        wsb = ord('\xe0')       # win32 MBS start mark

        # return keystrokes buffered by previous calls immediately,
        # regardless of timeout
        if len(self.i_buf):
            return self.i_buf.pop()

        # returns time-relative remaining for user-specified timeout
        timeleft = lambda cmp_time: (
            None if timeout is None else
            timeout - (time.time() - cmp_time)
            if timeout != 0 else 0)

        # returns True if byte appears to mark the beginning of a MBS
        chk_start = lambda char: (ord(char) == esc or (
            sys.platform == 'win32' and ord(char) == wsb))

        stime = time.time()
        ready = self.kbhit(timeleft(stime))
        if not ready:
            return None
        byte = self.getch()
        buf = [byte, ]
        # check for MSB; or just buffer for a rapidly firing input stream
        # TODO: meta sends escape", where alt+1 would send '\x1b1' may be imposed
        # a performance hit by the multibyte sequence decoder. It must be made
        # aware.
        while True:
            if chk_start(buf[0]):
                if (len(buf) == 1 and self.kbhit(esc_delay)) or self.kbhit():
                    byte = self.getch()
                    buf.append(byte)
                    detect = self.resolve_mbs(self._decode_istream(buf)).next()
                    if (detect.is_sequence and detect.code != self.KEY_EXIT):
                        # end of MBS,
                        break
                else:
                    # a poll for a 2nd+ byte failed; simple old ASCII escape.
                    break
            elif self.kbhit():
                # more still immediately available; buffer. this also catches
                # utf-8, I would suppose, as its not caught by chk_start().
                byte = self.getch()
                buf.append(byte)
            else:
                # no more bytes available in 'check for multibyte seq' loop
                break
        for keystroke in self.resolve_mbs(self._decode_istream(buf)):
            self.i_buf.insert(0, keystroke)
        item = self.i_buf.pop()
        return item

    @contextmanager
    def mouse_tracking(self, mode=MMODE_DEFAULT):
        """ Return a context manager for sending the DEC control sequence
        for entering the specified mouse tracking mode. The default mode is
        the urxvt extension found in xterm, urxvt, iTerm2, and gnome-terminal."""
        # this guy egmot made a push to get this mode in all utf-8 terminals,
        # http://www.midnight-commander.org/ticket/2662
        # http://www.xfree86.org/current/ctlseqs.html#Mouse%20Tracking
        # https://bugzilla.gnome.org/show_bug.cgi?id=662423
        # http://web.fis.unico.it/public/rxvt/refer.html#Mouse
        if mode == MMODE_XTERM_262:
            self.stream.write('\x1b[?%dh' % (MMODE_VT200,))
            self.stream.write('\x1b[?%dh' % (MMODE_EXT_MODE,))
        elif mode == MMODE_URXVT_910:
            self.stream.write('\x1b[?%dh' % (MMODE_VT200,))
            self.stream.write('\x1b[?%dh' % (MMODE_URXVT_EXT_MODE,))
        else:
            self.stream.write('\x1b[?%dh' % (mode,))
        self.stream.flush()
        try:
            yield
        finally:
            if mode in (MMODE_XTERM_262, MMODE_URXVT_910):
                self.stream.write('\x1b[?%dl' % (MMODE_VT200,))
            else:
                self.stream.write('\x1b[?%dl' % (mode,))


    @contextmanager
    def cbreak(self):
        """Return a context manager for entering 'cbreak' mode.

        This is anagolous to calling python's tty.setcbreak(), except
        that a finally clause restores the terminal state.

        In cbreak mode (sometimes called "rare" mode) normal tty line
        buffering is turned off and characters are available to be read
        one by one by ``getch()``. echo of input is also disabled, the
        application must explicitly copy-out any input received.

        More information can be found in the manual page for curses.h,
           http://www.openbsd.org/cgi-bin/man.cgi?query=cbreak
        Or the python manual for curses,
           http://docs.python.org/2/library/curses.html
        Note also that setcbreak sets VMIN = 1 and VTIME = 0,
           http://www.unixwiz.net/techtips/termios-vmin-vtime.html

        This is anagolous to the curses.wrapper() helper function, and
        the python standard module tty.setcbreak(), with the exception
        that the original terminal mode is restored when leaving context.
        """
        assert self.is_a_tty, u'stream is not a a tty.'
        mode = termios.tcgetattr(self.i_stream.fileno())
        tty.setcbreak(self.i_stream.fileno(), termios.TCSANOW)
        try:
            yield
        finally:
            termios.tcsetattr(self.i_stream.fileno(), termios.TCSAFLUSH, mode)

    @contextmanager
    def location(self, x=None, y=None):
        """Return a context manager for temporarily moving the cursor.

        Move the cursor to a certain position on entry, let you print stuff
        there, then return the cursor to its original position::

            term = Terminal()
            with term.location(2, 5):
                print 'Hello, world!'
                for x in xrange(10):
                    print 'I can do it %i times!' % x

        Specify ``x`` to move to a certain column, ``y`` to move to a certain
        row, both, or neither. If you specify neither, only the saving and
        restoration of cursor position will happen. This can be useful if you
        simply want to restore your place after doing some manual cursor
        movement.

        """
        # Save position and move to the requested column, row, or both:
        self.stream.write(self.save)
        if x is not None and y is not None:
            self.stream.write(self.move(y, x))
        elif x is not None:
            self.stream.write(self.move_x(x))
        elif y is not None:
            self.stream.write(self.move_y(y))
        yield

        # Restore original cursor position:
        self.stream.write(self.restore)

    @contextmanager
    def fullscreen(self):
        """Return a context manager that enters fullscreen mode while inside it and restores normal mode on leaving."""
        self.stream.write(self.enter_fullscreen)
        yield
        self.stream.write(self.exit_fullscreen)

    @contextmanager
    def hidden_cursor(self):
        """Return a context manager that hides the cursor while inside it and makes it visible on leaving."""
        self.stream.write(self.hide_cursor)
        yield
        self.stream.write(self.normal_cursor)

    @property
    def color(self):
        """Return a capability that sets the foreground color.

        The capability is unparametrized until called and passed a number
        (0-15), at which point it returns another string which represents a
        specific color change. This second string can further be called to
        color a piece of text and set everything back to normal afterward.

        :arg num: The number, 0-15, of the color

        """
        return ParametrizingString(self._foreground_color, self.normal)

    def wrap(self, ucs, width=None, **kwargs):
        """
        A.wrap(S, [width=None, indent=u'']) -> unicode

        Like textwrap.wrap, but honor existing linebreaks and understand
        printable length of a unicode string that contains ANSI sequences,
        such as colors, bold, etc. if width is not specified, the terminal
        width is used.
        """
        if width is None:
            width = self.width
        lines = []
        for line in ucs.splitlines():
            if line.strip():
                for wrapped in ansiwrap(line, width, **kwargs):
                    lines.append(wrapped)
            else:
                lines.append(u'')
        return lines

    @property
    def on_color(self):
        """Return a capability that sets the background color.

        See ``color()``.

        """
        return ParametrizingString(self._background_color, self.normal)

    @property
    def number_of_colors(self):
        """Return the number of colors the terminal supports.

        Common values are 0, 8, 16, 88, and 256.

        Though the underlying capability returns -1 when there is no color
        support, we return 0. This lets you test more Pythonically::

            if term.number_of_colors:
                ...

        We also return 0 if the terminal won't tell us how many colors it
        supports, which I think is rare.

        """
        # This is actually the only remotely useful numeric capability. We
        # don't name it after the underlying capability, because we deviate
        # slightly from its behavior, and we might someday wish to give direct
        # access to it.
        # Returns -1 if no color support, -2 if no such cap.
        colors = tigetnum('colors')
        # self.__dict__['colors'] = ret  # Cache it. It's not changing.
        # (Doesn't work.)
        return colors if colors >= 0 else 0

    def _resolve_formatter(self, attr):
        """Resolve a sugary or plain capability name, color, or compound formatting function name into a callable capability."""
        if attr in COLORS:
            return self._resolve_color(attr)
        elif attr in COMPOUNDABLES:
            # Bold, underline, or something that takes no parameters
            return self._formatting_string(self._resolve_capability(attr))
        else:
            formatters = split_into_formatters(attr)
            if all(f in COMPOUNDABLES for f in formatters):
                # It's a compound formatter, like "bold_green_on_red". Future
                # optimization: combine all formatting into a single escape
                # sequence.
                return self._formatting_string(
                    u''.join(self._resolve_formatter(s) for s in formatters))
            else:
                return ParametrizingString(self._resolve_capability(attr))

    def _resolve_capability(self, atom):
        """Return a terminal code for a capname or a sugary name, or an empty Unicode.

        The return value is always Unicode, because otherwise it is clumsy
        (especially in Python 3) to concatenate with real (Unicode) strings.

        """
        code = tigetstr(self._sugar.get(atom, atom))
        if code:
            # We can encode escape sequences as UTF-8 because they never
            # contain chars > 127, and UTF-8 never changes anything within that
            # range..
            return code.decode('utf-8')
        return u''

    def _decode_istream(self, buf, end=True):
        """ T._decode_istream(buf, end=True)

        Incrementaly decode input byte buffer, ``buf``, using the encoding
        specified by ``.encoding``. By default, encoding is the locale's
        preferred encoding detected into Unicode.
        """

        decoded = []
        for num, byte in enumerate(buf):
            is_final = end and num == (len(buf) - 1)
            ucs = self._idecoder.decode(byte, final=is_final)
            if ucs is not None:
                decoded.append(ucs)
        return u''.join(decoded)

    def resolve_mbs(self, ucs):
        """ T._resolve_mbs(ucs) -> Keystroke

        This generator yields unicode sequences with additional
        ``.is_sequence``, ``.name``, and ``.code`` properties that
        describle matching multibyte input sequences to keycode
        translations (if any) detected in input unicode buffer, ``ucs``.

        For win32 systems, the input buffer is a list of unicode values
        received by getwch. For Posix systems, the input buffer is a list
        of bytes recieved by sys.stdin.read(1), to be decoded to Unicode by
        the preferred locale.
        """
        CR_NVT = u'\r\x00'  # NVT return (telnet, etc.)
        CR_LF = u'\r\n'    # carriage return + newline
        CR_CHAR = u'\n'    # returns only '\n' when return is detected.
        esc = curses.ascii.ESC

        def resolve_keycode(integer):
            """
            Returns printable string to represent matched multibyte sequence,
            such as 'KEY_LEFT'. For purposes of __repr__ or __str__ ?
            """
            assert type(integer) is int
            for keycode in self._keycodes:
                if getattr(self, keycode) == integer:
                    return keycode

        def scan_keymap(ucs):
            """
            Return sequence and keycode if ucs begins with any known sequence.
            """
            for (keyseq, keycode) in self._keymap.iteritems():
                if ucs.startswith(keyseq):
                    return (keyseq, resolve_keycode(keycode), keycode)
            return (None, None, None)  # no match

        # special care is taken to pass over the ineveitably troublesome
        # carriage return, which is a multibyte sequence issue of its own;
        # expect to receieve any of '\r\00', '\r\n', '\r', or '\n', but
        # yield only a single byte, u'\n'.
        while len(ucs):
            if ucs[:2] in (CR_NVT, CR_LF):  # telnet return or dos CR+LF
                yield Keystroke(CR_CHAR, ('KEY_ENTER', self.KEY_ENTER))
                ucs = ucs[2:]
                continue
            elif ucs[:1] in (u'\r', u'\n'):  # single-byte CR
                yield Keystroke(CR_CHAR, ('KEY_ENTER', self.KEY_ENTER))
                ucs = ucs[1:]
                continue
            elif 1 == len(ucs) and ucs == unichr(esc):
                yield Keystroke(ucs[0], ('KEY_ESCAPE', self.KEY_ESCAPE))
                break
            elif 1 == len(ucs) and ucs == unichr(0):
                # null byte isn't; actually is ctrl+spacebar;
                yield Keystroke(ucs[0], ('KEY_CSPACE', self.KEY_CSPACE))
                break
            keyseq, keyname, keycode = scan_keymap(ucs)
            if (keyseq, keyname, keycode) == (None, None, None):
                yield Keystroke(ucs[0], None)
                ucs = ucs[1:]
            else:
                yield Keystroke(keyseq, (keyname, keycode))
                ucs = ucs[len(keyseq):]

    def _resolve_color(self, color):
        """Resolve a color like red or on_bright_green into a callable capability."""
        # TODO: Does curses automatically exchange red and blue and cyan and
        # yellow when a terminal supports setf/setb rather than setaf/setab?
        # I'll be blasted if I can find any documentation. The following
        # assumes it does.
        color_cap = (self._background_color if 'on_' in color else
                     self._foreground_color)
        # curses constants go up to only 7, so add an offset to get at the
        # bright colors at 8-15:
        offset = 8 if 'bright_' in color else 0
        base_color = color.rsplit('_', 1)[-1]
        return self._formatting_string(
            color_cap(getattr(curses, 'COLOR_' + base_color.upper()) + offset))

    @property
    def _foreground_color(self):
        return self.setaf or self.setf

    @property
    def _background_color(self):
        return self.setab or self.setb

    def _formatting_string(self, formatting):
        """Return a new ``FormattingString`` which implicitly receives my notion of "normal"."""
        return FormattingString(formatting, self.normal)

    def ljust(self, ucs, width=None, fillchar=u' '):
        """T.ljust(ucs, [width], [fillchar]) -> string

        Return ucs left-justified in a string of length width. Padding is
        done using the specified fill character (default is a space).
        Default width is terminal width. ucs is escape sequence safe."""
        if width is None:
            width = self.width
        return AnsiString(ucs).ljust(width, fillchar)

    def rjust(self, ucs, width=None, fillchar=u' '):
        """T.rjust(ucs, [width], [fillchar]) -> string

        Return ucs right-justified in a string of length width. Padding is
        done using the specified fill character (default is a space)
        Default width is terminal width. ucs is escape sequence safe."""
        if width is None:
            width = self.width
        return AnsiString(ucs).rjust(width, fillchar)

    def center(self, ucs, width=None, fillchar=u' '):
        """T.center(ucs, [width], [fillchar]) -> string

        Return ucs centered in a string of length width. Padding is
        done using the specified fill character (default is a space).
        Default width is terminal width. ucs is escape sequence safe."""
        if width is None:
            width = self.width
        return AnsiString(ucs).center(width, fillchar)


def derivative_colors(colors):
    """Return the names of valid color variants, given the base colors."""
    return set([('on_' + c) for c in colors] +
               [('bright_' + c) for c in colors] +
               [('on_bright_' + c) for c in colors])


COLORS = set(['black', 'red', 'green',
             'yellow', 'blue', 'magenta', 'cyan', 'white'])
COLORS.update(derivative_colors(COLORS))
COMPOUNDABLES = (COLORS |
                 set(['bold', 'underline', 'reverse', 'blink', 'dim', 'italic',
                      'shadow', 'standout', 'subscript', 'superscript']))


class ParametrizingString(unicode):
    """A Unicode string which can be called to parametrize it as a terminal capability"""
    def __new__(cls, formatting, normal=None):
        """Instantiate.

        :arg normal: If non-None, indicates that, once parametrized, this can
            be used as a ``FormattingString``. The value is used as the
            "normal" capability.

        """
        new = unicode.__new__(cls, formatting)
        new._normal = normal
        return new

    def __call__(self, *args):
        try:
            # Re-encode the cap, because tparm() takes a bytestring in Python
            # 3. However, appear to be a plain Unicode string otherwise so
            # concats work.
            parametrized = tparm(self.encode('utf-8'), *args).decode('utf-8')
            return (parametrized if self._normal is None else
                    FormattingString(parametrized, self._normal))
        except curses.error:
            # Catch "must call (at least) setupterm() first" errors, as when
            # running simply `nosetests` (without progressive) on nose-
            # progressive. Perhaps the terminal has gone away between calling
            # tigetstr and calling tparm.
            return u''
        except TypeError:
            # If the first non-int (i.e. incorrect) arg was a string, suggest
            # something intelligent:
            if len(args) >= 1 and isinstance(args[0], basestring):
                raise TypeError(
                    'A native or nonexistent capability template received '
                    '%r when it was expecting ints. You probably misspelled a '
                    'formatting call like bright_red_on_white(...).' % args)
            else:
                # Somebody passed a non-string; I don't feel confident
                # guessing what they were trying to do.
                raise


class FormattingString(unicode):
    """A Unicode string which can be called upon a piece of text to wrap it in formatting"""
    def __new__(cls, formatting, normal):
        new = unicode.__new__(cls, formatting)
        new._normal = normal
        return new

    def __call__(self, text):
        """Return a new string that is ``text`` formatted with my contents.

        At the beginning of the string, I prepend the formatting that is my
        contents. At the end, I append the "normal" sequence to set everything
        back to defaults. The return value is always a Unicode.

        """
        return self + text + self._normal


class NullCallableString(unicode):
    """A dummy class to stand in for ``FormattingString`` and ``ParametrizingString``

    A callable bytestring that returns an empty Unicode when called with an int
    and the arg otherwise. We use this when there is no tty and so all
    capabilities are blank.

    """
    def __new__(cls):
        new = unicode.__new__(cls, u'')
        return new

    def __call__(self, arg):
        if isinstance(arg, int):
            return u''
        return arg  # TODO: Force even strs in Python 2.x to be unicodes? Nah. How would I know what encoding to use to convert it?


class Keystroke(unicode):
    """ A unicode-derived class for matched multibyte input sequences.  If the
    unicode string is a multibyte input sequence, then the ``is_sequence``
    property is True, and the ``name`` and ``value`` properties return a
    string and integer value.
    """
    def __new__(cls, ucs, keystroke=None):
        new = unicode.__new__(cls, ucs)
        new._keystroke = keystroke
        return new

    @property
    def is_sequence(self):
        """ Returns True if value represents a multibyte sequence. """
        return self._keystroke is not None

    @property
    def name(self):
        """ Returns string name of multibyte sequence, such as 'KEY_HOME'."""
        if self._keystroke is None:
            return str(None)
        return self._keystroke[0]

    @property
    def code(self):
        """ Returns curses integer value of multibyte sequence, such as 323."""
        if self._keystroke is not None:
            return self._keystroke[1]
        assert 1 == len(self), (
            'No integer value available for multibyte sequence')
        return ord(self)


def split_into_formatters(compound):
    """Split a possibly compound format string into segments.

    >>> split_into_formatters('bold_underline_bright_blue_on_red')
    ['bold', 'underline', 'bright_blue', 'on_red']

    """
    merged_segs = []
    # These occur only as prefixes, so they can always be merged:
    mergeable_prefixes = ['on', 'bright', 'on_bright']
    for s in compound.split('_'):
        if merged_segs and merged_segs[-1] in mergeable_prefixes:
            merged_segs[-1] += '_' + s
        else:
            merged_segs.append(s)
    return merged_segs


class AnsiWrapper(textwrap.TextWrapper):
    # pylint: disable=C0111
    #         Missing docstring
    def _wrap_chunks(self, chunks):
        """
        ANSI-safe varient of wrap_chunks, with exception of movement seqs!
        """
        lines = []
        if self.width <= 0:
            raise ValueError("invalid width %r (must be > 0)" % self.width)
        chunks.reverse()
        while chunks:
            cur_line = []
            cur_len = 0
            if lines:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent
            width = self.width - len(indent)
            if (not hasattr(self, 'drop_whitespace') or self.drop_whitespace
                    ) and (chunks[-1].strip() == '' and lines):
                del chunks[-1]
            while chunks:
                chunk_len = len(AnsiString(chunks[-1]))
                if cur_len + chunk_len <= width:
                    cur_line.append(chunks.pop())
                    cur_len += chunk_len
                else:
                    break
            if chunks and len(AnsiString(chunks[-1])) > width:
                self._handle_long_word(chunks, cur_line, cur_len, width)
            if (not hasattr(self, 'drop_whitespace') or self.drop_whitespace
                    ) and (cur_line and cur_line[-1].strip() == ''):
                del cur_line[-1]
            if cur_line:
                lines.append(indent + u''.join(cur_line))
        return lines


def ansiwrap(ucs, width=70, **kwargs):
    """ Wrap a single paragraph of Unicode terminal sequences,
    returning a list of wrapped lines. ucs is ANSI-color safe.
    """
    assert ('break_long_words' not in kwargs
            or not kwargs['break_long_words']), (
                'break_long_words is not sequence-safe')
    kwargs['break_long_words'] = False
    return AnsiWrapper(width=width, **kwargs).wrap(ucs)

_ANSI_COLOR = re.compile(r'\033\[(\d{2,3})m')
_ANSI_RIGHT = re.compile(r'\033\[(\d{1,4})C')
_ANSI_CODEPAGE = re.compile(r'\033[\(\)][AB012]')
_ANSI_WILLMOVE = re.compile(r'\033\[[HuABCDEFG]')
_ANSI_WONTMOVE = re.compile(r'\033\[[smJK]')

class AnsiString(unicode):
    """
    This unicode variation understands the effect of ANSI sequences of
    printable length, allowing a properly implemented .rjust(), .ljust(),
    .center(), and .len() with bytes using terminal sequences

    Other ANSI helper functions also provided as methods.
    """
    # this is really bad; kludge dating as far back as 2002
    def __new__(cls, ucs):
        new = unicode.__new__(cls, ucs)
        return new

    def ljust(self, width, fillchar=u' '):
        return self + fillchar * (max(0, width - self.__len__()))

    def rjust(self, width, fillchar=u' '):
        return fillchar * (max(0, width - self.__len__())) + self

    def center(self, width, fillchar=u' '):
        split = max(0.0, float(width) - self.__len__()) / 2
        return (fillchar * (max(0, int(math.floor(split)))) + self
                + fillchar * (max(0, int(math.ceil(split)))))

    def __len__(self):
        """
        Return the printed length of a string that contains (some types) of
        ANSI sequences. Although accounted for, strings containing sequences
        such as cls() will not give accurate returns (0). backspace, delete,
        and double-wide east-asian characters are accounted for.
        """
        # 'nxt' points to first *ch beyond current ansi sequence, if any.
        # 'width' is currently estimated display length.
        nxt, width = 0, 0

        def get_padding(ucs):
            """
             get_padding(S) -> integer

            Returns int('nn') in CSI sequence \\033[nnC for use with replacing
            ansi.right(nn) with printable characters. prevents bleeding when
            used with scrollable art. Otherwise 0 if not \033[nnC sequence.
            Needed to determine the 'width' of art that contains this padding.
            """
            right = _ANSI_RIGHT.match(ucs)
            if right is not None:
                return int(right.group(1))
            return 0
        for idx in range(0, unicode.__len__(self)):
            width += get_padding(self[idx:])
            if idx == nxt:
                nxt = idx + _seqlen(self[idx:])
            if nxt <= idx:
                # 'East Asian Fullwidth' and 'East Asian Wide' characters
                # can take 2 cells, see http://www.unicode.org/reports/tr11/
                # and http://www.gossamer-threads.com/lists/python/bugs/972834
                # TODO: could use wcswidth, but i've ommitted it -jq
                width += 1
                nxt = idx + _seqlen(self[idx:]) + 1
        return width


def _is_movement(ucs):
    """
    _is_movement(ucs) -> bool

    Returns True if ucs begins with a terminal sequence that is
    "unhealthy for padding", that is, it has effects on the
    cursor position that are indeterminate.
    """
    # pylint: disable=R0911,R09120
    #        Too many return statements (20/6)
    #        Too many branches (23/12)
    # this isn't the best, perhaps for readability a giant REGEX can and
    # probably and already has been made.
    esc = curses.ascii.ESC
    slen = unicode.__len__(ucs)
    if 0 == slen:
        return False
    elif ucs[0] in u'\r\n\b':
        return True
    elif ucs[0] != unichr(esc):
        return False
    elif slen > 1 and ucs[1] == u'c':
        # reset
        return True
    elif slen < 3:
        # unknown
        return False
    elif _ANSI_CODEPAGE.match(ucs):
        return False
    elif (ucs[0], ucs[1], ucs[2]) == (u'\x1b', u'#', u'8'):
        # 'fill the screen'
        return True
    elif _ANSI_WILLMOVE.match(ucs):
        return True
    elif _ANSI_WONTMOVE.match(ucs):
        return False
    elif slen < 4:
        # unknown
        return False
    elif ucs[2] == '?':
        # DEC private modes
        # http://web.fis.unico.it/public/rxvt/refer.html#PrivateModes
        # fe., show/hide cursor is CSI + '?25[hl]'
        ptr2 = 3
        while (ucs[ptr2].isdigit()):
            ptr2 += 1
        if not ucs[ptr2] in u'tsrhl': # toggle,save,restore,set,reset
            return False
        return False
    elif ucs[2] in ('(', ')'):
        # CSI + '\([AB012]' # set G0/G1
        assert ucs[3] in (u'A', 'B', '0', '1', '2',)
        return False
    elif not ucs[2].isdigit():
        # illegal nondigit in seq
        return False
    ptr2 = 2
    while (ucs[ptr2].isdigit()):
        ptr2 += 1
    # multi-attribute SGR '[01;02(..)'(m|H)
    n_tries = 0
    while ptr2 < slen and ucs[ptr2] == ';' and n_tries < 64:
        n_tries += 1
        ptr2 += 1
        try:
            while (ucs[ptr2].isdigit()):
                ptr2 += 1
            if ucs[ptr2] == 'H':
                # 'H' pos,
                return True
            elif ucs[ptr2] == 'm':
                # 'm' color;attr
                return False
            elif ucs[ptr2] == ';':
                # multi-attribute SGR
                continue
        except IndexError:
            # out-of-range in multi-attribute SGR
            return False
        # illegal multi-attribtue SGR
        return False
    if ptr2 >= slen:
        # unfinished sequence, hrm ..
        return False
    elif ucs[ptr2] in u'ABCD':
        # numeric up, down, right, left
        return True
    elif ucs[ptr2] in u'KJ':
        # clear_bol is 1K; clear_eos is 2J
        return False
    elif ucs[ptr2] == 'm':
        # normal
        return False
    # illegal single value, UNKNOWN
    return False


def _seqlen(ucs):
    """
    _seqlen(S) -> integer

    Returns non-zero for string S that begins with an ansi sequence, with
    value of bytes until sequence is complete. Use as a 'next' pointer to
    skip past sequences.
    """
    # pylint: disable=R0911,R0912
    #        Too many return statements (19/6)
    #        Too many branches (22/12)
    # it is regretable that this duplicates much of is_movement, but
    # they do serve different means .. again, more REGEX would help
    # readability.
    slen = unicode.__len__(ucs)
    esc = curses.ascii.ESC
    if 0 == slen:
        return 0  # empty string
    elif ucs[0] != unichr(esc):
        return 0  # not a sequence
    elif 1 == slen:
        return 0  # just esc,
    elif ucs[1] == u'c':
        return 2  # reset
    elif 2 == slen:
        return 0  # not a sequence
    elif (ucs[1], ucs[2]) == (u'#', u'8'):
        return 3  # fill screen (DEC)
    elif _ANSI_CODEPAGE.match(ucs) or _ANSI_WONTMOVE.match(ucs):
        return 3
    elif _ANSI_WILLMOVE.match(ucs):
        return 3
    elif ucs[1] == '[':
        # all sequences are at least 4 (\033,[,0,m)
        if slen < 4:
            # not a sequence !?
            return 0
        elif ucs[2] == '?':
            # CSI + '?25(h|l)' # show|hide
            ptr2 = 3
            while (ucs[ptr2].isdigit()):
                ptr2 += 1
                if (ucs[ptr2]) == ';':
                    # u'\x1b[?12;25h' cvvis
                    ptr2 += 1
            if not ucs[ptr2] in u'hl':
                # ? followed illegaly, UNKNOWN
                return 0
            return ptr2 + 1
        # SGR
        elif ucs[2].isdigit() or ucs[2] == ';':
            ptr2 = 2
            while (ucs[ptr2].isdigit()):
                ptr2 += 1
                if ptr2 == unicode.__len__(ucs):
                    return 0

            # multi-attribute SGR '[01;02(..)'(m|H)
            while ucs[ptr2] == ';':
                ptr2 += 1
                if ptr2 == unicode.__len__(ucs):
                    return 0
                try:
                    while (ucs[ptr2].isdigit()):
                        ptr2 += 1
                except IndexError:
                    return 0
                if ucs[ptr2] in u'Hm':
                    return ptr2 + 1
                elif ucs[ptr2] == ';':
                    # multi-attribute SGR
                    continue
                # 'illegal multi-attribute sgr'
                return 0
            # single attribute SGT '[01(A|B|etc)'
            if ucs[ptr2] in u'ABCDEFGJKSTHm':
                # single attribute,
                # up/down/right/left/bnl/bpl,pos,cls,cl,
                # pgup,pgdown,color,attribute.
                return ptr2 + 1
            # illegal single value
            return 0
        # illegal nondigit
        return 0
    # unknown...
    return 0



"""
Session script for x/84, https://github.com/jquast/x84

This script displays a CP437 artwork (block ansi), and prompts the user to
chose 'utf8' or 'cp437' encoding. Other than the default, 'utf8', the
Session.write() method takes special handling of a session.encoding value
of 'cp437' for encoding translation.

feedback appreciated in the refresh() method; special characters
are used to "switch" terminals from one mode to another -- do any work?
"""

import os


def get_selector(selection):
    """
    Instantiate a new selector, dynamicly for the window size.
    """
    from x84.bbs import getterminal, Selector
    term = getterminal()
    width = max(30, (term.width / 2) - 10)
    xloc = max(0, (term.width / 2) - (width / 2))
    sel = Selector(yloc=term.height - 1,
                   xloc=xloc, width=width,
                   left='utf8', right='cp437')
    sel.selection = selection
    return sel


def main():
    """ Main procedure. """
    # pylint: disable=R0912
    #        Too many branches
    from x84.bbs import getsession, getterminal, echo, getch, Ansi, from_cp437
    session, term = getsession(), getterminal()
    session.activity = u'Selecting chracter set'
    artfile = os.path.join(
        os.path.dirname(__file__), 'art', (
            'plant-256.ans' if term.number_of_colors == 256
            else 'plant.ans'))
    enc_prompt = (
        u'Press left/right until artwork looks best. Clients should'
        ' select utf8 encoding and Andale Mono font. Older clients or'
        ' clients with appropriate 8-bit fontsets can select cp437, though'
        ' some characters may appear as "?".')
    save_msg = u"\r\n\r\n'%s' is now your preferred encoding ..\r\n"
    if session.user.get('expert', False):
        echo(u'\r\n\r\n(U) UTF-8 encoding or (C) CP437 encoding [uc] ?\b\b')
        while True:
            inp = getch()
            if inp in (u'u', u'U'):
                session.encoding = 'utf8'
                break
            elif inp in (u'c', u'C'):
                session.encoding = 'cp437'
                break
        session.user['charset'] = session.encoding
        echo(save_msg % (session.encoding,))
        getch(1.0)
        return

    art = (from_cp437(open(artfile).read()).splitlines()
           if os.path.exists(artfile) else [u''])

    def refresh(sel):
        """ Refresh art and yes/no prompt, ``sel``. """
        session.flush_event('refresh')
        session.encoding = selector.selection
        if sel.selection == 'utf8':
            # ESC %G activates UTF-8 with an unspecified implementation
            # level from ISO 2022 in a way that allows to go back to
            # ISO 2022 again.
            echo(unichr(27) + u'%G')
        elif sel.selection == 'cp437':
            # ESC %@ returns to ISO 2022 in case UTF-8 had been entered.
            # ESC ) U Sets character set G1 to codepage 437 .. usually.
            echo(unichr(27) + u'%@')
            echo(unichr(27) + u')U')
        else:
            assert False, "Only encodings 'utf8' and 'cp437' supported."
        # display art, banner, paragraph, refresh selector refresh
        buf = [line for line in art]
        return u''.join((
            u'\r\n\r\n',
            u'\r\n'.join(buf),
            u'\r\n\r\n',
            Ansi(enc_prompt).wrap(int(term.width * .95)),
            u'\r\n\r\n',
            sel.refresh(),))

    selector = get_selector(session.encoding)
    echo(refresh(selector))
    while True:
        inp = getch(1)
        if inp == term.KEY_ENTER:
            session.user['charset'] = session.encoding
            echo(save_msg % (session.encoding,))
            getch(1.0)
            return
        elif inp is not None:
            selector.process_keystroke(inp)
            if selector.quit:
                # 'escape' quits without save, though the encoding
                # has been temporarily set for this session.
                return
            if selector.moved:
                # set and refresh art in new encoding
                echo(refresh(selector))
        if session.poll_event('refresh') is not None:
            # instantiate a new selector in case the window size has changed.
            selector = get_selector(session.encoding)
            echo(refresh(selector))



""" Last Callers script for x/84, http://github.com/jquast/x84 """
# this also allows viewing of '.plan' attribute string when set by user,
# or 'e'diting a user when executed by sysop -- gosub('profile', user)


def pak():
    """ Press any key prompt. """
    from x84.bbs import echo, getch
    msg_pak = u'PRESS ANY kEY'
    echo(u'\r\n%s ... ' % (msg_pak,))
    getch()
    return


def view_plan(handle):
    """ Display .plan file for handle. """
    from x84.bbs import getterminal, echo, Ansi, get_user
    term = getterminal()
    echo(u'\r\n\r\n')
    echo(Ansi(get_user(handle).get('.plan', u'No Plan.')).wrap(term.width))
    echo(u'\r\n')
    pak()


def dummy_pager(last_callers):
    """ Dummy pager for displaying last callers """
    # pylint: disable=R0914
    #         Too many local variables
    from x84.bbs import getterminal, getsession, echo, getch, ini
    from x84.bbs import LineEditor, Ansi, list_users, get_user, gosub
    session, term = getsession(), getterminal()
    msg_prompt = (
        u'\r\n%sONtiNUE, %stOP, %sON-StOP %siEW .PlAN%s ?\b\b' % (
        term.bold(u'[c]'),
        term.bold(u'[s]'),
        term.bold(u'n'),
        term.bold(u'[v]'),
        u' [e]dit USR' if (
        'sysop' in session.user.groups) else u'',))
    msg_partial = u'PARtiAl MAtChES'
    msg_prompt_handle = u'ViEW .PlAN ::- ENtER hANdlE: '

    redraw()
    echo(u'\r\n\r\n')
    nonstop = False
    row = 10  # after-art,
    for txt in last_callers:
        echo(Ansi(txt).ljust(term.width / 2).center(term.width))
        echo(u'\r\n')
        row += 1
        if ((not nonstop and row > 0 and 0 == (row % (term.height - 3)))
                or (row == len(last_callers) - 1)):
            echo(msg_prompt)
            inp = getch()
            row = 2
            if inp in (u's', u'S', u'q', u'Q', term.KEY_EXIT):
                return
            if inp in (u'v', u'V') or 'sysop' in session.user.groups and (
                    inp in (u'e', u'E')):
                echo(u'\r\n\r\n')
                echo(msg_prompt_handle)
                handle = LineEditor(ini.CFG.getint('nua', 'max_user')).read()
                usrlist = list_users()
                if handle is None or 0 == len(handle.strip()):
                    continue
                handle = handle.strip()
                if handle.lower() in [nick.lower() for nick in list_users()]:
                    user = get_user((nick for nick in usrlist
                                     if nick.lower() == handle.lower()).next())
                    if 'sysop' in session.user.groups and (
                            inp in (u'e', u'E')):
                        gosub('profile', user.handle)
                    else:
                        view_plan(user.handle)
                else:
                    misses = [nick for nick in usrlist.keys()
                              if nick.lower().startswith(handle[:1].lower())]
                    if len(misses) > 0:
                        echo(u'%s:\r\n\r\n%s\r\n' % (msg_partial,
                            Ansi(', '.join(misses)).wrap(term.width)))
                    continue
            if inp in ('n', u'N'):
                nonstop = True
            echo(u'\r\n\r\n')
    pak()


def refresh_opts(pager, handle):
    """ Refresh pager border with command keys available. """
    from x84.bbs import getsession, getterminal, get_user, find_user, Ansi
    session, term = getsession(), getterminal()
    if not handle or not find_user(handle):
        has_plan = 0
    else:
        has_plan = 0 != len(get_user(handle).get('.plan', u'').strip())
    decorate = lambda key, desc: u''.join((
        term.red_underline(key,),
        u':',
        term.yellow(desc.split()[0]), u' ',
        u' '.join(desc.split()[1:]),
        u' ' if len(desc.split()) > 1 else u'',))
    statusline = u''.join((
        term.bold_yellow(u'- '),
        decorate(u'Escape/q', 'Uit'),
        decorate(u'v', 'iEW .PLAN') if has_plan else u'',
        decorate(u'e', 'dit USR') if 'sysop' in session.user.groups else u'',
        term.bold_yellow(u'-'),
    ))
    if len(Ansi(statusline)) < (pager.visible_width - 4):
        return pager.border() + pager.footer(statusline)
    else:
        return pager.border() + pager.footer(term.bold_red('q') + u':uit')


def get_lightbar(lcallers, lcalls):
    """
    Return UI element for browsing last callers, given ``lcallers`` as
    a list of handles, and parallel array ``lcalls`` as unicode string
    to display for last call of each handle.
    """
    from x84.bbs import getterminal, Lightbar
    term = getterminal()
    width = min(50, max(10, term.width - 5))
    height = max(4, min(term.height - 8, 35))
    xloc = (term.width / 2) - (width / 2)
    yloc = term.height - height
    pager = Lightbar(height, width, yloc, xloc)
    pager.glyphs['left-vert'] = pager.glyphs['right-vert'] = u''
    pager.colors['highlight'] = term.red_reverse
    pager.colors['border'] = term.yellow
    pager.xpadding, pager.ypadding = 2, 1
    pager.alignment = 'center'
    pager.update([(lcallers[n], txt,)
                  for (n, txt) in enumerate(lcalls.split('\n'))])
    return pager


def get_art(fname):
    """ Return ansi art center-aligned. """
    from x84.bbs import getterminal
    term = getterminal()
    buf = list()
    width = 0
    for line in open(fname):
        art_line = line.rstrip()[:term.width - 1]
        width = max(len(art_line), width)
        buf.append(art_line)
    return [line.center(width) for line in buf]


def redraw(pager=None):
    """ Returns unicode sequence suitable for redrawing screen. """
    from x84.bbs import getterminal
    import os
    term = getterminal()
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'lc.asc')
    # return banner, artfile contents, pager refreshed
    return u''.join((
        u'\r\n\r\n',
        term.bold_red(u'/'.rjust(
            pager.xloc if pager is not None else term.width / 3)),
        term.red(u'/'), u'l', term.red('AS'),
        term.bold_red('t '), u'C', term.red('A'),
        term.bold_red('ll'), term.red('ERS'),
        u'\r\n',
        u'\r\n'.join(get_art(artfile)),
        u'\r\n',
        u'\r\n' * (pager.height if pager is not None else 0),
        pager.border() if pager is not None else u'',
        pager.refresh() if pager is not None else u'',))


def lc_retrieve():
    """
    Returns tuple of ([nicknames,] u'text'), where 'text' describes in Ansi
    color the last callers to the system, and 'nicknames' is simply a list
    of last callers (for lightbar selection key).
    """
    # pylint: disable=R0914
    #         Too many local variables
    from x84.bbs import get_user, ini, timeago, getterminal
    from x84.bbs import DBProxy
    import time
    term = getterminal()
    udb = DBProxy('lastcalls')
    # re-order by time called; unfortunate ..; note that sqlite
    # encodes unicode as utf-8; but doesn't decode it on retrieval,
    # of dict keys; possible upstream patching opportunity here,
    sortdb = {}
    for ((handle), (tm_lc, _nc, origin)) in (udb.items()):
        while tm_lc in sortdb:
            tm_lc += 0.1
        sortdb[tm_lc] = [handle.decode('utf-8'), _nc, origin]

    padd_handle = (ini.CFG.getint('nua', 'max_user') + 2)
    padd_origin = (ini.CFG.getint('nua', 'max_location') + 2)
    rstr = u''
    nicks = []
    for tm_lc, (handle, _nc, origin) in (reversed(sorted(sortdb.items()))):
        is_sysop = 'sysop' in get_user(handle).groups
        rstr += (term.bold_red(u'@') if is_sysop else u''
                 ) + (term.ljust(handle,
                     (padd_handle - (2 if is_sysop else 1))))
        rstr += term.red(origin.ljust(padd_origin))
        rstr += timeago(time.time() - tm_lc)
        rstr += u'\n'
        nicks.append(handle)
    return (nicks, rstr.rstrip())


def main():
    """ Main procedure. """
    from x84.bbs import getsession, getterminal, echo, getch, gosub
    session, term = getsession(), getterminal()
    lcallers, lcalls_txt = lc_retrieve()
    lbr = None
    dirty = True
    handle = None

    if (0 == term.number_of_colors
            or session.user.get('expert', False)):
        echo(redraw(None))
        return dummy_pager(lcalls_txt.split('\n'))

    while lbr is None or not lbr.quit:
        if dirty or lbr is None or session.poll_event('refresh'):
            session.activity = u'Viewing last callers'
            lcallers, lcalls_txt = lc_retrieve()
            pos = lbr.position if lbr is not None else (0, 0)
            lbr = get_lightbar(lcallers, lcalls_txt)
            if pos:
                lbr.position = pos
            echo(redraw(lbr))
            echo(refresh_opts(lbr, handle))
        sel = lbr.selection[0]
        if sel != handle or dirty:
            handle = sel
            echo(refresh_opts(lbr, handle))
            echo(lbr.pos(lbr.yloc + (lbr.height - 1)))
            dirty = False
            continue
        inp = getch(1)
        if inp is not None:
            if inp in (u'v', u'V'):
                view_plan(handle)
                dirty = True
            elif inp in (u'e', u'E') and 'sysop' in session.user.groups:
                gosub('profile', handle)
                dirty = True
            else:
                echo(lbr.process_keystroke(inp))



""" Debug example script for x/84, https://github.com/jquast/x84 """


def main():
    """ Main procedure. """
    # by default, nothing is done.
    from x84.bbs import getsession
    assert 'sysop' in getsession().user.groups

    return migrate_105lc()
    #return nothing()

    # but this is a great way to make data manipulations,
    # exampled here is importing of a .csv import of
    # a mystic recordbase.
    #return merge_mystic()

    #return tygerofdantye_fix()


def migrate_105lc():
    from x84.bbs import echo, DBProxy, list_users, get_user
    # migrating lastcallers database for 1.0.5 upgrade
    lc = DBProxy('lastcalls')
    for handle in list_users():
        user = get_user(handle)
        lc[(handle)] = (user.lastcall, user.calls, user.location)
        echo('\r\n' + user.handle + '.')
    echo ('\r\n\r\nlast callers db rebuilt!')


def nothing():
    """ Do nothing. """
    from x84.bbs import echo, getch
    echo(u'Nothing to do.')
    getch(3)

def tygerofdantye_fix():
    """ This user was too long! """
    from x84.bbs import DBProxy
    user = DBProxy('userbase')['tygerofdanyte']
    user.delete()
    user.handle = u'tygerdanyte'
    user.save()


def merge_mystic():
    """ Example script to merge csv records into userbase. """
    # pylint: disable=R0914
    #         Too many local variables
    from x84.bbs import ini, echo, getch, User, get_user, find_user
    import os
    # you must modify variable ``do_write`` to commit changes,
    # csv format; 'user:pass:origin:email\n', in iso8859-1 encoding.
    do_write = False
    inp_file = os.path.join(
        ini.CFG.get('system', 'datapath'),
        'mystic_dat.csv')
    lno = 0
    for lno, line in enumerate(open(inp_file, 'r')):
        handle = line.split(':', 1)[0].strip().decode('iso8859-1')
        attrs = line.rstrip().split(':')[2:]
        (_password, _location, _email) = attrs
        (_password, _location, _email) = (
            _password.strip().decode('iso8859-1'),
            _location.strip().decode('iso8859-1'),
            _email.strip().decode('iso8859-1'))
        echo(u''.join((u'\r\n',
                       handle, u': ',
                       '%d ' % (len(_password)),
                       '%s ' % (_location),
                       '%s ' % (_email),)))
        match = find_user(handle)
        if match is None:
            user = User(handle)
            user.location = _location
            user.email = _email
            user.password = _password
        else:
            user = get_user(match)
        user.groups.add('old-school')
        if do_write:
            user.save()
    echo('\r\n\r\n%d lines processed.' % (lno,))
    getch()



"""
Custom exception classes for x/84, https://github.com/jquast/x84
"""


class Disconnected(Exception):
    """Thrown when a client is disconnected."""
    pass


class Goto (Exception):
    """Thrown to change script without returning."""
    pass


class ScriptError (Exception):
    """Thrown for internal scripting errors."""
    pass



""" Password reset script for x/84, http://github.com/jquast/x84 """

from x84.bbs import getsession, getterminal, list_users, echo, getch, ini
from x84.bbs import LineEditor, get_user

from x84.default.nua import set_password
import logging
import smtplib
import base64
import os
from email.mime.text import MIMEText


def main(handle):
    """ Main procedure. """
    # pylint: disable=R0914,R0915,R0911
    #         Too many local variables
    #         Too many statements
    #         Too many return statements
    # by request from midget; a password reset form
    session, term = getsession(), getterminal()
    session.activity = 'resetting password for %s' % (handle,)
    user = get_user(handle)
    logger = logging.getLogger()

    prompt_email = u'\r\n\r\nENtER E-MAil fOR %r: '
    msg_nfound = u'\r\n\r\n%r NOt fOUNd iN USERbASE.'
    msg_cancelled = u'\r\n CANCEllEd.'
    msg_wrong = u'\r\n\r\nWRONG'
    msg_mailsubj = u'passkey for %s' % (ini.CFG.get('system', 'bbsname'))
    msg_mailbody = u'Your passkey is %r'
    msg_mailfrom = ini.CFG.get('system', 'mail_addr')
    msg_sent = u'\r\n\r\nPASSkEY hAS bEEN SENt tO %s.'
    prompt_passkey = u'\r\n\r\nENtER PASSkEY: '
    msg_verified = u'\r\n\r\nYOU hAVE bEEN VERifiEd.'

    echo(term.normal)
    if not handle in list_users():
        echo(term.bold_red(msg_nfound))
        getch()
        return False

    width = ini.CFG.getint('nua', 'max_email')
    email = None
    tries = 0
    while True:
        tries += 1
        if tries > 5:
            logger.warn('%r email retries exceeded', handle)
            return False
        echo(prompt_email % (handle,))
        try_email = LineEditor(width).read()
        if try_email is None or 0 == len(try_email):
            echo(term.normal + msg_cancelled)
            return False

        # fetch user record email
        email = get_user(handle).email
        if email is None or 0 == len(email):
            logger.warn('%r missing email address, cannot send', handle)
            echo(term.bold_red(msg_wrong))

        elif email.lower() != try_email.lower():
            logger.warn('%r failed email %r (try: %r)', handle,
                        email, try_email)
            echo(term.bold_red(msg_wrong))

        else:
            logger.info('%r requests password reset to %r', handle, email)
            break

    # generate a 'passkey' and e-mail out of band, and request input
    passkey = base64.encodestring(
        os.urandom(ini.CFG.getint('nua', 'max_pass'))
    )[:ini.CFG.getint('nua', 'max_pass')]
    msg = MIMEText(msg_mailbody % (passkey,))
    msg['From'] = msg_mailfrom
    msg['To'] = email
    msg['Subject'] = msg_mailsubj
    # pylint: disable=W0703
    #         Catching too general exception Exception
    err = None
    try:
        smtp = smtplib.SMTP(ini.CFG.get('system', 'mail_smtphost'))
        smtp.sendmail(msg_mailfrom, [email], msg.as_string())
        smtp.quit()
    except Exception as err:
        logger.exception(err)
        echo('u\r\n\r\n' + term.bold_red(str(err)) + u'\r\n')
        getch(2)
        return False

    echo(msg_sent % (email,))
    width = len(passkey)
    email = None
    tries = 0
    while True:
        tries += 1
        if tries > 5:
            logger.warn("passkey retries exceeded for user '%s'", handle)
            return False

        echo(term.normal + u'\r\n\r\n')
        echo(prompt_passkey)
        try_passkey = LineEditor(width).read()
        if try_passkey is None or 0 == len(try_passkey):
            echo(term.normal + msg_cancelled)
            logger.warn("cancelled passkey for user '%s'", handle)
            return False

        if passkey == try_passkey:
            echo(term.bold_green(msg_verified))
            echo(u'\r\n\r\n')
            break
        logger.warn("failed passkey for user '%s': '%s', tried '%s')",
                    handle, passkey, try_passkey)
        echo(term.bold_red(msg_wrong))

    set_password(user)
    user.save()
    return True



"""
editor package for x/84, https://github.com/jquast/x84
"""

from x84.bbs.ansiwin import AnsiWindow
from x84.bbs.output import Ansi

PC_KEYSET = {'refresh': [unichr(12), ],
             'backspace': [unichr(8), unichr(127), ],
             'backword': [unichr(23), ],
             'enter': [u'\r', ],
             'exit': [unichr(27), ], }


class LineEditor(object):
    """
    This unicode line editor is unaware of its (x, y) position and is great
    for prompting a quick phrase on any terminal, such as a login: prompt.
    """
    # This should really be gnu/readline, but its not really accessible ..
    _hidden = False
    _width = 0

    def __init__(self, width=None, content=u''):
        """
        Arguments:
            width: the maximum input length
        """
        self._width = width
        self.content = content
        self._quit = False
        self._carriage_returned = False
        self.colors = dict()
        self.keyset = PC_KEYSET
        self.init_keystrokes()
        self.init_theme()

    def init_theme(self):
        """
        Initialize colors['highlight'].
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        self.colors['highlight'] = term.reverse

    def init_keystrokes(self):
        """
        This initializer sets keyboard keys for backspace/exit.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        self.keyset['refresh'].append(term.KEY_REFRESH)
        self.keyset['backspace'].append(term.KEY_BACKSPACE)
        self.keyset['enter'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_EXIT)

    @property
    def quit(self):
        """
        Returns: True if a terminating or quit character was handled by
        process_keystroke(), such as the escape key, or 'q' by default.
        """
        return self._quit

    @property
    def carriage_returned(self):
        """
        Returns True when last keystroke caused carriage to be returned.
        (KEY_ENTER was pressed)
        """
        return self._carriage_returned

    @property
    def hidden(self):
        """
        When not False, represents a single 'mask' character to hide input
        with, such as a password prompt
        """
        return self._hidden

    @hidden.setter
    def hidden(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert value is False or 1 == len(value)
        self._hidden = value

    @property
    def width(self):
        """
        When non-zero, represents the upperbound limit of characters to receive
        on input until no more characters are accepted.
        """
        return self._width

    @width.setter
    def width(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._width = value

    def refresh(self):
        """
        Returns unicode byts suitable for drawing line.
        No movement or positional sequences are returned.
        """
        from x84.bbs.session import getterminal
        lightbar = u''.join((
            self.colors.get('highlight', u''),
            ' ' * self.width,
            '\b' * self.width))
        content = (self.hidden * len(Ansi(self.content))
                   if self.hidden else self.content)
        return u''.join((
            lightbar,
            content,
            getterminal().cursor_visible))

    def process_keystroke(self, keystroke):
        """
        Process the keystroke received by read method and take action.
        """
        self._quit = False
        if keystroke in self.keyset['refresh']:
            return u'\b' * len(Ansi(self.content)) + self.refresh()
        elif keystroke in self.keyset['backspace']:
            if len(self.content) != 0:
                len_toss = len(Ansi(self.content[-1]))
                self.content = self.content[:-1]
                return u''.join((
                    u'\b' * len_toss,
                    u' ' * len_toss,
                    u'\b',))
        elif keystroke in self.keyset['backword']:
            if len(self.content) != 0:
                ridx = self.content.rstrip().rfind(' ') + 1
                toss = len(Ansi(self.content[ridx:]))
                move = len(self.content[ridx:])
                self.content = self.content[:ridx]
                return u''.join((
                    u'\b' * toss,
                    u' ' * move,
                    u'\b' * move,))
        elif keystroke in self.keyset['enter']:
            self._carriage_returned = True
        elif keystroke in self.keyset['exit']:
            self._quit = True
        elif type(keystroke) is int:
            return u''
        elif (ord(keystroke) >= ord(' ') and
                (len(Ansi(self.content)) < self.width or self.width == 0)):
            self.content += keystroke
            return keystroke if not self.hidden else self.hidden
        return u''

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).
        Allows backspacing. Returns unicode text, or None when cancelled.
        """
        from x84.bbs import getch
        from x84.bbs.output import echo
        from x84.bbs.session import getsession, getterminal
        session, term = getsession(), getterminal()
        self._carriage_returned = False
        self._quit = False
        echo(self.refresh())
        while not (self.quit or self.carriage_returned):
            inp = getch()
            echo(self.process_keystroke(inp))
        echo(term.normal)
        if not self.quit:
            return self.content
        return None


class ScrollingEditor(AnsiWindow):
    """
    A single line Editor fully aware of its (y, x).
    Infinite horizontal scrolling is enabled with the enable_scrolling
    property. When True, scrolling amount is limited using max_length.
    """
    # pylint: disable=R0902,R0904
    #        Too many instance attributes (14/7)
    #        Too many public methods (33/20)

    def __init__(self, width, yloc, xloc):
        """
        Construct a Line editor at (y,x) location of width (n).
        """
        self._horiz_shift = 0
        self._horiz_pos = 0
        self._enable_scrolling = False
        self._horiz_lastshift = 0
        self._scroll_pct = 35.0
        self._margin_pct = 20.0
        self._carriage_returned = False
        self._max_length = 0
        self._quit = False
        self._bell = False
        self.content = u''
        self.keyset = PC_KEYSET
        height = 3  # TODO: 2 of 3 for top and bottom border
        # (optionaly displayed .. is this best x/y coord?
        #   once working, lets set default as borderless!)
        AnsiWindow.__init__(self, height, width, yloc, xloc)
        self.init_keystrokes()
        self.init_theme()

    __init__.__doc__ = AnsiWindow.__init__.__doc__

    @property
    def position(self):
        """
        Tuple of shift amount and column position of line editor.
        """
        return (self._horiz_shift, self._horiz_pos)

    @property
    def eol(self):
        """
        Return True when no more input can be accepted (end of line).
        """
        return len(Ansi(self.content)) >= self.max_length

    @property
    def bell(self):
        """
        Returns True when user nears margin and bell has been sounded and
        carriage has not yet been returned.
        """
        return int(float(self.visible_width) * (float(self.scroll_pct) * .01))

    @bell.setter
    def bell(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._bell = value

    @property
    def carriage_returned(self):
        """
        Returns True when last keystroke caused carriage to be returned.
        (KEY_ENTER was pressed)
        """
        return self._carriage_returned

    @property
    def quit(self):
        """
        Returns: True if a terminating or quit character was handled by
        process_keystroke(), such as the escape key, or 'q' by default.
        """
        return self._quit

    @property
    def enable_scrolling(self):
        """
        Enable horizontal scrolling of line editor.
        Otherwise, input is limited to visible width.
        """
        return self._enable_scrolling

    @enable_scrolling.setter
    def enable_scrolling(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._enable_scrolling = value

    @property
    def is_scrolled(self):
        """
        Returns True if the horizontal editor is in a scrolled state.
        """
        return self._horiz_shift > 0

    @property
    def scroll_amt(self):
        """
        Returns number of columns horizontal editor will scroll, calculated by
        scroll_pct.
        """
        return int(float(self.visible_width) * (float(self.scroll_pct) * .01))

    @property
    def margin_amt(self):
        """
        Returns number of columns from right-edge that the horizontal editor
        signals bell=True, indicating that the end is near and the carriage
        should be soon returned. This also indicates the distance from margin
        a user may type into until ScrollingEditor horizontally shifts.
        """
        return int(float(self.visible_width) * (float(self.margin_pct) * .01))

    @property
    def scroll_pct(self):
        """
        Number of columns, as a percentage of its total visible width, that
        will be scrolled when a user reaches the margin and enable_scrolling
        is True. Default is 35.
        """
        return self._scroll_pct

    @scroll_pct.setter
    def scroll_pct(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._scroll_pct = float(value)

    @property
    def margin_pct(self):
        """
        Number of columns away from input length limit, as a percentage of its
        total visible width, that will alarm the bell. This simulates the bell
        of a typewriter as a signaling mechanism. Default is 20.
        """
        # .. unofficially; intended to be used be a faked multi-line editor, by
        # using the bell as a wrap signal to instantiate another line editor
        # and 'return the carriage'
        return self._margin_pct

    @margin_pct.setter
    def margin_pct(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._margin_pct = float(value)

    @property
    def max_length(self):
        """
        Maximum line length. This also limits infinite scrolling when
        enable_scrolling is True. When unset, the maximum length is the
        visible width of the window.
        """
        return self._max_length or self.visible_width

    @max_length.setter
    def max_length(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._max_length = value

    def init_theme(self):
        """
        Initialize colors['highlight'] as REVERSE and glyphs['strip'] as '$ '.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        AnsiWindow.init_theme(self)
        self.colors['highlight'] = term.yellow_reverse
        self.glyphs['strip'] = '$ '

    def init_keystrokes(self):
        """
        This initializer sets glyphs and colors appropriate for a "theme",
        override or inherit this method to create a common color and graphic
        set.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        self.keyset['refresh'].append(term.KEY_REFRESH)
        self.keyset['backspace'].append(term.KEY_BACKSPACE)
        self.keyset['enter'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_EXIT)

    def process_keystroke(self, keystroke):
        """
        Process the keystroke received by read method and take action.
        """
        self._quit = False
        rstr = u''
        if keystroke in self.keyset['refresh']:
            rstr = self.refresh()
        elif keystroke in self.keyset['backspace']:
            rstr = self.backspace()
        elif keystroke in self.keyset['backword']:
            rstr = self.backword()
        elif keystroke in self.keyset['enter']:
            self._carriage_returned = True
            rstr = u''
        elif keystroke in self.keyset['exit']:
            self._quit = True
            rstr = u''
        elif type(keystroke) is int:
            rstr = u''
        else:
            rstr = self.add(keystroke)
        return rstr

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).
        Allows backspacing. Returns unicode text, or None when cancelled.
        """
        from x84.bbs import getch
        from x84.bbs.output import echo
        echo(self.refresh())
        self._quit = False
        self._carriage_returned = False
        while not (self.quit or self.carriage_returned):
            inp = getch()
            echo(self.process_keystroke(inp))
        if not self.quit:
            return self.content
        return None

    def fixate(self, x_adjust=0):
        """
        Return terminal sequence suitable for placing cursor at current
        position in window. Set x_adjust to -1 to position cursor 'on'
        the last character, or 0 for 'after' (default).
        """
        from x84.bbs.session import getterminal
        xpos = self._xpadding + self._horiz_pos + x_adjust
        return self.pos(1, xpos) + getterminal().cursor_visible

    def refresh(self):
        """
        Return unicode sequence suitable for refreshing the entire
        line and placing the cursor.

        A strange by-product; if scrolling was not previously enabled,
        it is if wrapping must occur; this can happen if a
        non-scrolling editor was provided a very large .content
        buffer, then later .refresh()'d. -- essentially enabling
        infinate scrolling
        """
        # reset position and detect new position
        self._horiz_lastshift = self._horiz_shift
        self._horiz_shift = 0
        self._horiz_pos = 0
        for _count in range(len(Ansi(self.content))):
            if (self._horiz_pos >
                    (self.visible_width - self.scroll_amt)):
                self._horiz_shift += self.scroll_amt
                self._horiz_pos -= self.scroll_amt
                self.enable_scrolling = True
            self._horiz_pos += 1
        if self._horiz_shift > 0:
            self._horiz_shift += len(self.glyphs['strip'])
            # self._horiz_shift += len(self.glyphs['strip'])
            prnt = self.glyphs['strip'] + self.content[self._horiz_shift:]
        else:
            prnt = self.content
        erase_len = self.visible_width - (len(prnt))
        return u''.join((
            self.pos(self.ypadding, self.xpadding),
            self.colors.get('highlight', u''),
            prnt + self.glyphs.get('erase', u' ') * erase_len,
            self.fixate(),))

    def backword(self):
        """
        Delete word behind cursor, using ' ' as boundary.
        in readline this is unix-word-rubout (C-w).
        """
        if 0 == len(self.content):
            return u''
        ridx = self.content.rstrip().rfind(' ') + 1
        self.content = self.content[:ridx]
        return self.refresh()

    def backspace(self):
        """
        Remove character from end of content buffer,
        scroll as necessary.
        """
        if 0 == len(self.content):
            return u''
        rstr = u''
        # measured backspace erases over double-wide
        len_toss = len(Ansi(self.content[-1]))
        len_move = len(self.content[-1])
        self.content = self.content[:-1]
        if (self.is_scrolled and (self._horiz_pos < self.scroll_amt)):
            # shift left,
            self._horiz_shift -= self.scroll_amt
            self._horiz_pos += self.scroll_amt
            rstr += self.refresh()
        else:
            rstr += u''.join((
                self.fixate(0),
                u'\b' * len_toss,
                u' ' * len_move,
                u'\b' * len_move,))
        self._horiz_pos -= 1
        return rstr

    def update(self, ucs=u''):
        """
        Replace or reset content.
        """
        self._horiz_shift = 0
        self._horiz_pos = 0
        self.content = ucs
        assert unichr(27) not in ucs, ('Editor is not ESC sequence-safe')
        # TODO: assert binary, also ..

    def add(self, u_chr):
        """
        Returns output sequence necessary to add a character to
        content buffer.  An empty content buffer is returned if
        no data could be inserted. Sequences for re-displaying
        the full input line are returned when the character
        addition caused the window to scroll horizontally.
        Otherwise, the input is simply returned to be displayed
        ('local echo').
        """
        from x84.bbs import getterminal
        term = getterminal()
        if self.eol:
            # cannot input, at end of line!
            return u''
        # append to input
        self.content += u_chr
        # return character appended as output, ensure .fixate() is used first!
        self._horiz_pos += 1
        if self._horiz_pos >= (self.visible_width - self.margin_amt):
            # scrolling is required,
            return self.refresh()
        return term.normal + self.colors['highlight'] + u_chr



""" ttyplay door for x/84, https://github.com/jquast/x84 """
import os

def playfile(ttyplay_exe, ttyfile, peek=False):
    from x84.bbs import getsession, getterminal, echo, getch, Door
    import re
    session, term = getsession(), getterminal()

    if not (os.path.exists(ttyfile) and os.path.isfile(ttyfile)):
        echo(term.bold_red('\r\n\r\nPAth NOt fOUNd: %s\r\n' % (ttyfile,)))
        getch()
        return

    # .. we could look for and write various information headers..
    # .. esp. at EOF, about session.user.handle & connect_time & etc.
    data = open(ttyfile, 'rb').read(64)
    size_pattern = re.compile(r'\[8;(\d+);(\d+)t')
    match = size_pattern.findall(data)
    if match:
        echo(u'\r\n\r\nheight, width: %s, %s' % match[0].groups())
    args = tuple()
    if peek:
        args += ('-p',)
    elif 'sysop' in session.user.groups:
        echo("\r\nPRESS '%s' tO PEEk (2)\b\b" % (
            (term.green_underline(u'p'),)))
        if getch(2) in (u'p', u'P'):
            peek = True
    args += (ttyfile,)
    door = Door(ttyplay_exe, args=args)
    session.activity = u'playing tty recording'
    # press any key prompt, instructions for quitting (^c) ..
    echo(u'\r\n\r\n lOAd CASSEttE ANd PRESS %s. PRESS %s tO %s.\r\n' % (
        term.green(u'PlAY'),
        term.red_reverse('^c'),
        term.red_underline(u'StOP'),))
    getch(2)
    with term.fullscreen():
        door.run()
        echo(u'\r\n' * term.height)
        echo(u'PRESS ' + term.green_underline('q') + u'.')
        while not getch() in (u'q', u'Q'):
            pass

def main(ttyfile=u'', peek=False):
    """ Main procedure. """
    from x84.bbs import Lightbar, getch, getsession, getterminal, ini, echo
    # pylint: disable=R0914,R0915
    #         Too many local variables
    #         Too many statements
    ttyplay_exe = ini.CFG.get('ttyplay', 'exe')
    if not os.path.exists(ttyplay_exe):
        echo(u'\r\n%s NOt iNStAllEd.\r\n' % (ttyplay_exe,))
        getch()
        return
    session, term = getsession(), getterminal()
    # pylint: disable=W0212
    #         Access to a protected member _record_tty of a client class
    resume_rec = session._record_tty and session.is_recording
    if resume_rec:
        session.stop_recording()
        session._record_tty = False

    if 'sysop' in session.user.groups and ttyfile == u'':
        # pylint: disable=W0212
        #         Access to a protected member _ttyrec_folder of a client class
        folder = os.path.dirname(ttyfile) or session._ttyrec_folder
        pos = None
        while True:
            files = sorted([fn for fn in os.listdir(session._ttyrec_folder)
                            if fn.endswith('.ttyrec')])
            echo(u'\r\n' * term.height)
            sel = Lightbar(term.height - 1, term.width - 1, 0, 0)
            sel.colors['border'] = term.bold_green
            echo(sel.border() + sel.title('-  SElECt A RECORdiNG  -'))
            sel.update([(fname, fname) for fname in files])
            if pos is not None:
                sel.position = pos

            x_ttyfile = sel.read()
            if x_ttyfile is None or sel.quit:
                break
            pos = sel.position
            ttyfile = os.path.join(folder, x_ttyfile)
            playfile(ttyplay_exe, ttyfile, peek)
    else:
        playfile(ttyplay_exe, ttyfile, peek)

    if not session.is_recording and resume_rec:
        session._record_tty = True
        session.start_recording()

    echo(term.move(term.height, 0))
    echo(u'\r\n')




"""
 Matrix login screen for X/84 (Formerly, 'The Progressive') BBS,

 This script is the session entry point.  Or simply put, the login program.

 In legacy era, a matrix script might be something to fool folk not in the
 know, meant to divert agents from underground boards, require a passcode,
 or even swapping the modem into a strange stop/bit/parity configuration,
 callback mechanisms, etc.. read all about it in old e-zines.
"""

import os


def denied(msg):
    """ Display denied message, pause for input for 1s. """
    from x84.bbs import getterminal, echo, getch
    term = getterminal()
    echo(u'\r\n' + term.bold_red(msg))
    getch(1.0)


def get_username(handle=u''):
    """
    Prompt for a login handle. If unfound, script change to 'nua' when
    allow_apply is enabled (default=yes). Also allow 'anonymous' when enabled
    (default=no). A unicode handle of non-zero length is returned when the
    login handle matches a userbase record.
    """
    # pylint: disable=R0914,R0911
    #         Too many local variables
    #         Too many return statements
    from x84.bbs import getterminal, ini, echo, LineEditor, gosub, goto
    from x84.bbs import find_user, getch
    term = getterminal()
    prompt_user = u'\r\n  user: '
    apply_msg = u'\r\n\r\n  --> Create new account? [ynq]   <--' + '\b' * 5
    allow_apply = ini.CFG.getboolean('nua', 'allow_apply')
    enable_anonymous = ini.CFG.getboolean('matrix', 'enable_anonymous')
    # pylint: disable=E1103
    #         Instance of '_Chainmap' has no 'split' member
    #         (but some types could not be inferred)
    newcmds = ini.CFG.get('matrix', 'newcmds').split()
    byecmds = ini.CFG.get('matrix', 'byecmds').split()
    denied_msg = u'\r\n\r\nfiRSt, YOU MUSt AbANdON YOUR libERtIES.'
    badanon_msg = u"\r\n  " + term.bright_red + u"'%s' login denied."
    max_user = ini.CFG.getint('nua', 'max_user')
    nuascript = ini.CFG.get('nua', 'script')

    echo(prompt_user)
    handle = LineEditor(max_user, handle).read()
    if handle is None or 0 == len(handle.strip()):
        echo(u'\r\n')
        return u''
    elif handle.lower() in newcmds:
        if allow_apply:
            gosub('nua', u'')
            return u''
        denied(denied_msg)
        return u''
    elif handle.lower() in byecmds:
        goto('logoff')
    elif handle.lower() == u'anonymous':
        if enable_anonymous:
            goto(topscript, 'anonymous')
        denied(badanon_msg % (handle,))
        return u''
    u_handle = find_user(handle)
    if u_handle is not None:
        return u_handle  # matched
    if allow_apply is False:
        denied(denied_msg)
        return u''

    echo(apply_msg)
    ynq = getch()
    if ynq in (u'q', u'Q', term.KEY_EXIT):
        # goodbye
        goto('logoff')
    elif ynq in (u'y', u'Y'):
        # new user application
        goto(nuascript, handle)
    echo(u'\r\n')
    return u''


def try_reset(user):
    """ Prompt for password reset. """
    from x84.bbs import echo, getch, gosub
    prompt_reset = u'RESEt PASSWORD (bY E-MAil)? [yn]'
    echo(prompt_reset)
    while True:
        inp = getch()
        if inp in (u'y', u'Y'):
            return gosub('pwreset', user.handle)
        elif inp in (u'n', u'N'):
            echo(u'\r\n\r\n')
            return False


def try_pass(user):
    """
    Prompt for password and authenticate, returns True if succesfull.
    """
    # pylint: disable=R0914
    #         Too many local variables
    from x84.bbs import getsession, getterminal, ini, LineEditor, echo
    session, term = getsession(), getterminal()
    prompt_pass = u'\r\n\r\n  pass: '
    status_auth = u'\r\n\r\n  ' + term.yellow_reverse(u"Encrypting ..")
    badpass_msg = (u'\r\n\r\n' + term.red_reverse +
                   u"'%s' login failed." + term.normal)
    max_pass = int(ini.CFG.get('nua', 'max_pass'))
    # prompt for password, disable input tap during, mask input with 'x',
    # and authenticate against user record, performing a script change to
    # topscript if sucessful.
    # pylint: disable=W0212
    #         Access to a protected member _tap_input of a client class
    echo(prompt_pass)
    chk = session._tap_input  # <-- save
    session._tap_input = False
    lne = LineEditor(max_pass)
    lne.hidden = u'x'
    password = lne.read()
    session._tap_input = chk  # restore -->
    if password is not None and 0 != len(password):
        echo(status_auth)
        if user.auth(password):
            return True
    denied(badpass_msg % (user.handle,))
    return False


def uname():
    """
    On unix systems with uname, call with -a on connect
    """
    from x84.bbs import Door
    for uname_filepath in ('/usr/bin/uname', '/bin/uname'):
        if os.path.exists(uname_filepath):
            Door(uname_filepath, args=('-a',)).run()
            break


def main():
    """ Main procedure. """
    # pylint: disable=R0914,R0911
    #         Too many local variables
    import logging
    from x84.bbs import getsession, getterminal, ini, echo, get_user, goto
    from x84.bbs import find_user
    from x84.engine import __url__ as url
    logger = logging.getLogger()
    session, term = getsession(), getterminal()
    session.activity = u'Logging in'
    handle = (session.env.get('USER', '').decode('iso8859-1', 'replace'))
    anon_allowed_msg = u"'%s' login enabled.\r\n" % (
        term.bold_cyan('anonymous',))
    # pylint: disable=E1103
    #         Instance of '_Chainmap' has no 'split' member
    #         (but some types could not be inferred)
    newcmds = ini.CFG.get('matrix', 'newcmds').split()
    apply_msg = u"'%s' to create new account.\r\n" % (
        term.bold_cyan(newcmds[0]),)
    allow_apply = ini.CFG.getboolean('nua', 'allow_apply')
    enable_anonymous = ini.CFG.getboolean('matrix', 'enable_anonymous')
    enable_pwreset = ini.CFG.getboolean('matrix', 'enable_pwreset')
    bbsname = ini.CFG.get('system', 'bbsname')
    artfile = os.path.join(os.path.dirname(__file__), 'art', '1984.asc')
    topscript = ini.CFG.get('matrix', 'topscript')
    max_tries = 10
    session.flush_event('refresh')
    #uname()
    # display banner
    echo(u''.join((
        term.normal, u'\r\n',
        u'Connected to %s, see %s for source\r\n' % (bbsname, url),
        u'\r\n'.join([line[:term.width].rstrip() for line in open(artfile)]),
        u'\r\n\r\n',
        term.bold(u'tERM'), u': ',
        term.cyan_underline(session.env['TERM']),
        u'\r\n',
        term.bold(u'diMENSiONs'), u': ', '%s%s%s' % (
            term.bold_cyan(str(term.width)),
            term.cyan(u'x'),
            term.bold_cyan(str(term.height)),),
        u'\r\n',
        term.bold(u'ENCOdiNG'), u': ',
        term.cyan_underline(session.encoding),
        u'\r\n\r\n',
        anon_allowed_msg if enable_anonymous else u'',
        apply_msg if allow_apply else u'',
    )))
    # http://www.termsys.demon.co.uk/vtansi.htm
    # disable line-wrapping
    echo(unichr(27) + u'[7l')

    # http://www.xfree86.org/4.5.0/ctlseqs.html
    # Save xterm icon and window title on stack.
    echo(unichr(27) + u'[22;0t')

    if handle:
        echo('\r\nHello, %s!' % (handle,))
        match = find_user(handle)
        if match is not None:
            handle = match
        else:
            handle = ''

    # prompt for username & password
    for _num in range(0, max_tries):
        handle = get_username(handle)
        if handle != u'':
            session.activity = u'Logging in'
            user = get_user(handle)
            if try_pass(user):
                goto(topscript, user.handle)
            echo(u'\r\n\r\n')
            if enable_pwreset:
                try_reset(user)
            else:
                logger.info('%r failed password', handle)
    logger.warn('maximum tries exceeded')
    goto('logoff')



""" news script for x/84, https://github.com/jquast/x84 """
NEWS_ART = None
NEWSAGE = 0
NEWS = None


def dummy_pager(news_txt):
    """
    Given news_txt as unicode string, display using a dummy pager.
    """
    from x84.bbs import getterminal, echo, getch
    term = getterminal()
    prompt_msg = u'\r\n[%s]ontinue, [%s]top, [%s]on-stop  ?\b\b' % (
        term.bold_blue('c'), term.bold_blue('s'), term.bold_blue('n'),)
    nonstop = False
    echo(redraw(None))
    for row in range(len(news_txt)):
        echo(news_txt[row].rstrip() + '\r\n')
        if not nonstop and row > 0 and 0 == (row % (term.height - 3)):
            echo(prompt_msg)
            inp = getch()
            if inp in (u's', u'S', u'q', u'Q', term.KEY_EXIT):
                return
            if inp in ('n', u'N'):
                nonstop = True
            echo(u'\r\n')
    echo('\r\npress any key .. ')
    getch()
    return


def get_pager(news_txt, position=None):
    """
    Return Pager instance with content ``news_txt``.
    """
    from x84.bbs import getterminal, Pager
    term = getterminal()
    width = min(130, (term.width - 2))
    height = term.height - len(redraw(None).splitlines())
    yloc = term.height - height
    xloc = (term.width / 2) - (width / 2)
    pager = Pager(height, width, yloc, xloc)
    pager.colors['border'] = term.blue
    pager.glyphs['left-vert'] = u''
    pager.glyphs['right-vert'] = u''
    pager.update('\n'.join(news_txt))
    if position is not None:
        pager.position = position
    return pager


def redraw(pager):
    """ Returns string suitable for refreshing screen. """
    from x84.bbs import getsession, getterminal
    import os
    # pylint: disable=W0603
    #         Using the global statement
    global NEWS_ART  # in-memory cache
    session, term = getsession(), getterminal()
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'news.asc')
    if NEWS_ART is None:
        NEWS_ART = [line for line in open(artfile)]
    # left-align, center, strip, and trim each line of ascii art
    ladjust = lambda line: (
        line.rstrip().center(term.width)[:term.width].rstrip())
    title = u''.join((u']- ', term.bold_blue('PARtY NEWS'), ' [-',))
    footer = u''.join((u'-[ ',
                       term.blue_underline(u'Escape'), '/',
                       term.blue_underline(u'q'), term.bold_blue(u'uit '),
                       ((u'- ' + term.blue_underline(u'e')
                           + term.bold_blue(u'dit ')) if (
                               'sysop' in session.user.groups) else u''),
                           u']-',
    ))
    return u''.join((u'\r\n\r\n',
                     '\r\n'.join(
                         (ladjust(line) for line in NEWS_ART)), u'\r\n',
                     u''.join((
                              u'\r\n' * pager.height,
                              pager.refresh(),
                              pager.border(),
                              pager.title(title),
                              pager.footer(footer),)
                              ) if pager is not None else u'',))


def main():
    """ Main procedure. """
    from x84.bbs import getsession, echo, getch, gosub
    import codecs
    import time
    import os
    # pylint: disable=W0603
    #         Using the global statement
    global NEWS, NEWSAGE  # in-memory cache
    session = getsession()
    session.activity = 'Reading news'
    newsfile = os.path.join(os.path.dirname(__file__), 'art', 'news.txt')
    if not os.path.exists(newsfile):
        echo(u'\r\n\r\nNo news.')
        return

    pager = None
    dirty = True
    while True:
        if session.poll_event('refresh'):
            dirty = True
        if dirty:
            if NEWS is None or os.stat(newsfile).st_mtime > NEWSAGE:
                # open a utf-8 file for international encodings/art/language
                NEWSAGE = time.time()
                NEWS = [line.rstrip() for line in
                        codecs.open(newsfile, 'rb', 'utf8')]
            if (session.user.get('expert', False)):
                return dummy_pager(NEWS)
            pos = pager.position if pager is not None else None
            pager = get_pager(NEWS, pos)
            echo(redraw(pager))
            dirty = False
        inp = getch(1)
        if inp is not None:
            if inp in (u'e', u'E',) and 'sysop' in session.user.groups:
                session.user['news'] = u'\r\n'.join(NEWS)
                if gosub('editor', 'news'):
                    NEWS = session.user['news'].splitlines()
                    codecs.open(newsfile, 'wb', 'utf8').write(
                        u'\r\n'.join(NEWS))
                dirty = True
            else:
                echo(pager.process_keystroke(inp))
                if pager.quit:
                    return
